import json
import logging
from collections.abc import Iterable
from dataclasses import dataclass
from functools import partial

from databricks.sdk.errors import DatabricksError
from databricks.labs.blueprint.parallel import Threads
from databricks.labs.lsql.backends import SqlBackend
from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.mapping import TableMapping
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationStatusRefresher
from databricks.labs.ucx.recon.base import (
    DataComparator,
    SchemaComparator,
    TableIdentifier,
    DataComparisonResult,
)

logger = logging.getLogger(__name__)


@dataclass
class ReconResult:
    src_schema: str
    src_table: str
    dst_catalog: str
    dst_schema: str
    dst_table: str
    schema_matches: bool = False
    data_matches: bool = False
    schema_comparison: str | None = None
    data_comparison: str | None = None
    error_message: str | None = None


class MigrationRecon(CrawlerBase[ReconResult]):
    def __init__(
        self,
        sbe: SqlBackend,
        schema: str,
        migration_status_refresher: TableMigrationStatusRefresher,
        table_mapping: TableMapping,
        schema_comparator: SchemaComparator,
        data_comparator: DataComparator,
        default_threshold: float,
    ):
        super().__init__(sbe, "hive_metastore", schema, "recon_results", ReconResult)
        self._migration_status_refresher = migration_status_refresher
        self._table_mapping = table_mapping
        self._schema_comparator = schema_comparator
        self._data_comparator = data_comparator
        self._default_threshold = default_threshold

    def _crawl(self) -> Iterable[ReconResult]:
        self._migration_status_refresher.reset()
        migration_index = self._migration_status_refresher.index()
        migration_rules = self._table_mapping.load()
        tasks = []
        for source in migration_index.snapshot():
            migration_status = migration_index.get(*source)
            if migration_status is None:
                continue
            if not migration_status.dst_catalog or not migration_status.dst_schema or not migration_status.dst_table:
                continue
            source_table = TableIdentifier(
                "hive_metastore",
                migration_status.src_schema,
                migration_status.src_table,
            )
            target_table = TableIdentifier(
                migration_status.dst_catalog,
                migration_status.dst_schema,
                migration_status.dst_table,
            )
            compare_rows = False
            recon_tolerance_percent = self._default_threshold
            for rule in migration_rules:
                if rule.match(source_table):
                    compare_rows = rule.compare_rows
                    recon_tolerance_percent = rule.recon_tolerance_percent
                    break
            tasks.append(
                partial(
                    self._recon,
                    source_table,
                    target_table,
                    compare_rows,
                    recon_tolerance_percent,
                )
            )
        recon_results, errors = Threads.gather("reconciling data", tasks)
        if len(errors) > 0:
            logger.error(f"Detected {len(errors)} while reconciling data")
        return recon_results

    def _recon(
        self,
        source: TableIdentifier,
        target: TableIdentifier,
        compare_rows: bool,
        recon_tolerance_percent: int,
    ) -> ReconResult | None:
        try:
            schema_comparison = self._schema_comparator.compare_schema(source, target)
            data_comparison = self._data_comparator.compare_data(source, target, compare_rows)
        except DatabricksError as e:
            logger.warning(f"Error while comparing {source.fqn_escaped} and {target.fqn_escaped}: {e}")
            return ReconResult(
                source.schema,
                source.table,
                target.catalog,
                target.schema,
                target.table,
                error_message=str(e),
            )
        recon_results = ReconResult(
            source.schema,
            source.table,
            target.catalog,
            target.schema,
            target.table,
            schema_comparison.is_matching,
            (
                self._percentage_difference(data_comparison) <= recon_tolerance_percent
                and data_comparison.target_missing_count == 0
                and data_comparison.source_missing_count == 0
            ),
            json.dumps(schema_comparison.as_dict()),
            json.dumps(data_comparison.as_dict()),
        )
        return recon_results

    @staticmethod
    def _percentage_difference(result: DataComparisonResult) -> int:
        source_row_count = result.source_row_count
        target_row_count = result.target_row_count
        if source_row_count == 0 and target_row_count == 0:
            return 0
        if source_row_count == 0 and target_row_count > 0:
            return 100
        return round(abs(source_row_count - target_row_count) / source_row_count * 100)

    def _try_fetch(self) -> Iterable[ReconResult]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield ReconResult(*row)



import logging
import re
from collections.abc import Collection, Callable
from dataclasses import dataclass
from functools import partial

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.parallel import Threads
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import BadRequest, NotFound, ResourceConflict, DatabricksError, InvalidParameterValue
from databricks.sdk.service.catalog import TableInfo, SchemaInfo

from databricks.labs.ucx.account.workspaces import WorkspaceInfo
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore import TablesCrawler
from databricks.labs.ucx.hive_metastore.tables import Table
from databricks.labs.ucx.recon.base import TableIdentifier

logger = logging.getLogger(__name__)


@dataclass
class Rule:
    workspace_name: str
    catalog_name: str
    src_schema: str
    dst_schema: str
    src_table: str
    dst_table: str
    recon_tolerance_percent: int = 0  # threshold for row count comparison
    compare_rows: bool = False  # whether to compare row by row

    @classmethod
    def initial(cls, workspace_name: str, catalog_name: str, table: Table, recon_tolerance_percent: int) -> "Rule":
        return cls(
            workspace_name=workspace_name,
            catalog_name=catalog_name,
            src_schema=table.database,
            dst_schema=table.database,
            src_table=table.name,
            dst_table=table.name,
            recon_tolerance_percent=recon_tolerance_percent,
        )

    @classmethod
    def from_src_dst(cls, src_table: TableInfo, dst_schema: SchemaInfo) -> "Rule":
        return cls(
            workspace_name="workspace",
            catalog_name=str(dst_schema.catalog_name or ""),
            src_schema=str(src_table.schema_name or ""),
            dst_schema=str(dst_schema.name or ""),
            src_table=str(src_table.name or ""),
            dst_table=str(src_table.name or ""),
            recon_tolerance_percent=0,
        )

    def match(self, table: TableIdentifier) -> bool:
        return table.catalog == "hive_metastore" and self.src_schema == table.schema and self.src_table == table.table

    @property
    def as_uc_table_key(self) -> str:
        return f"{self.catalog_name}.{self.dst_schema}.{self.dst_table}"

    @property
    def as_hms_table_key(self) -> str:
        return f"hive_metastore.{self.src_schema}.{self.src_table}"


@dataclass
class TableToMigrate:
    src: Table
    rule: Rule

    def __hash__(self):
        return hash(self.src)

    def __eq__(self, other):
        return isinstance(other, TableToMigrate) and self.src == other.src


class TableMapping:
    FILENAME = 'mapping.csv'
    UCX_SKIP_PROPERTY = "databricks.labs.ucx.skip"

    def __init__(
        self,
        installation: Installation,
        ws: WorkspaceClient,
        sql_backend: SqlBackend,
        recon_tolerance_percent: int = 0,
    ):
        self._installation = installation
        self._ws = ws
        self._sql_backend = sql_backend
        self._recon_tolerance_percent = recon_tolerance_percent

    def current_tables(self, tables: TablesCrawler, workspace_name: str, catalog_name: str):
        tables_snapshot = list(tables.snapshot())
        if not tables_snapshot:
            msg = "No tables found. Please run: databricks labs ucx ensure-assessment-run"
            raise ValueError(msg)
        for table in tables_snapshot:
            yield Rule.initial(workspace_name, catalog_name, table, self._recon_tolerance_percent)

    def save(self, tables: TablesCrawler, workspace_info: WorkspaceInfo) -> str:
        workspace_name = workspace_info.current()
        default_catalog_name = re.sub(r"\W+", "_", workspace_name)
        current_tables = self.current_tables(tables, workspace_name, default_catalog_name)
        return self._installation.save(list(current_tables), filename=self.FILENAME)

    def load(self) -> list[Rule]:
        try:
            return self._installation.load(list[Rule], filename=self.FILENAME)
        except NotFound:
            msg = "Please run: databricks labs ucx create-table-mapping"
            raise ValueError(msg) from None

    def skip_table_or_view(self, schema_name: str, table_name: str, load_table: Callable[[str, str], Table | None]):
        # Marks a table to be skipped in the migration process by applying a table property
        try:
            table = load_table(schema_name, table_name)
            if table is None:
                raise NotFound("[TABLE_OR_VIEW_NOT_FOUND]")
            self._sql_backend.execute(
                f"ALTER {table.kind} {escape_sql_identifier(schema_name)}.{escape_sql_identifier(table_name)} SET TBLPROPERTIES('{self.UCX_SKIP_PROPERTY}' = true)"
            )
        except NotFound as err:
            if "[TABLE_OR_VIEW_NOT_FOUND]" in str(err) or "[DELTA_TABLE_NOT_FOUND]" in str(err):
                logger.error(f"Failed to apply skip marker for Table {schema_name}.{table_name}. Table not found.")
            else:
                logger.error(
                    f"Failed to apply skip marker for Table {schema_name}.{table_name}: {err!s}", exc_info=True
                )
        except BadRequest as err:
            logger.error(f"Failed to apply skip marker for Table {schema_name}.{table_name}: {err!s}", exc_info=True)

    def unskip_table_or_view(
        self, schema_name: str, table_name: str, load_table: Callable[[str, str], Table | None]
    ) -> None:
        """Removes skip mark from the table property.

        Args:
            schema_name (str): The schema name of the table to be unskipped.
            table_name (str): The table name of the table to be unskipped.
            load_table (Callable[[str, str], Table | None]): A function that loads a table from the metastore.
        """
        table = load_table(schema_name, table_name)
        if table is None:
            logger.error(
                f"Failed to remove skip marker from table: {schema_name}.{table_name}. Table not found.",
            )
            return
        try:
            self._sql_backend.execute(
                f"ALTER {table.kind} {escape_sql_identifier(table.full_name)} UNSET TBLPROPERTIES IF EXISTS('{self.UCX_SKIP_PROPERTY}');"
            )
        except (NotFound, BadRequest) as e:
            logger.error(f"Failed to remove skip marker from table: {table.full_name}", exc_info=e)

    def skip_schema(self, schema: str):
        # Marks a schema to be skipped in the migration process by applying a table property
        try:
            self._sql_backend.execute(
                f"ALTER SCHEMA {escape_sql_identifier(schema)} SET DBPROPERTIES('{self.UCX_SKIP_PROPERTY}' = true)"
            )
        except NotFound as err:
            if "[SCHEMA_NOT_FOUND]" in str(err):
                logger.error(f"Failed to apply skip marker for Schema {schema}. Schema not found.")
            else:
                logger.error(err)
        except BadRequest as err:
            logger.error(err)

    def unskip_schema(self, schema: str) -> None:
        """Removes skip mark from the schema property.

        Args:
            schema (str): The schema name of the table to be unskipped.
        """
        try:
            self._sql_backend.execute(
                f"ALTER SCHEMA hive_metastore.{escape_sql_identifier(schema)} UNSET DBPROPERTIES IF EXISTS('{self.UCX_SKIP_PROPERTY}');"
            )
        except (NotFound, BadRequest) as e:
            logger.error(f"Failed to remove skip marker from schema: {schema}.", exc_info=e)

    def get_tables_to_migrate(self, tables_crawler: TablesCrawler) -> Collection[TableToMigrate]:
        rules = self.load()
        # Getting all the source tables from the rules
        databases_in_scope = self._get_databases_in_scope({rule.src_schema for rule in rules})
        crawled_tables_keys = {crawled_table.key: crawled_table for crawled_table in tables_crawler.snapshot()}
        tasks = []
        for rule in rules:
            if rule.as_hms_table_key not in crawled_tables_keys:
                logger.info(f"Table {rule.as_hms_table_key} in the mapping doesn't show up in assessment")
                continue
            if rule.src_schema not in databases_in_scope:
                logger.info(f"Table {rule.as_hms_table_key} is in a database that was marked to be skipped")
                continue
            if crawled_tables_keys[rule.as_hms_table_key].is_databricks_dataset:
                logger.info(f"Table {rule.as_hms_table_key} is a db demo dataset and will not be migrated")
                continue
            tasks.append(
                partial(self._get_table_in_scope_task, TableToMigrate(crawled_tables_keys[rule.as_hms_table_key], rule))
            )

        return Threads.strict("checking all database properties", tasks)

    def _get_databases_in_scope(self, databases: set[str]):
        tasks = []
        for database in databases:
            tasks.append(partial(self._get_database_in_scope_task, database))
        return Threads.strict("checking databases for skip property", tasks)

    def _get_database_in_scope_task(self, database: str) -> str | None:
        if database.startswith("mounted_"):
            return database
        describe = {}
        try:
            for value in self._sql_backend.fetch(f"DESCRIBE SCHEMA EXTENDED {escape_sql_identifier(database)}"):
                describe[value["database_description_item"]] = value["database_description_value"]
        except NotFound:
            logger.warning(
                f"Schema hive_metastore.{database} no longer exists. Skipping its properties check and migration."
            )
            return None
        properties = describe.get("Properties", "")
        if not properties:
            return database
        if self.UCX_SKIP_PROPERTY in TablesCrawler.parse_database_props(properties.lower()):
            logger.info(f"Database {database} is marked to be skipped")
            return None
        return database

    def _get_table_in_scope_task(self, table_to_migrate: TableToMigrate) -> TableToMigrate | None:
        table = table_to_migrate.src
        rule = table_to_migrate.rule

        if self.exists_in_uc(table, rule.as_uc_table_key):
            logger.info(f"The intended target for {table.key}, {rule.as_uc_table_key}, already exists.")
            return None
        properties = self._get_table_properties(table)
        if not properties:
            return None

        for value in properties:
            if value["key"] == self.UCX_SKIP_PROPERTY:
                logger.info(f"{table.key} is marked to be skipped")
                return None
            if value["key"] == "upgraded_to":
                logger.info(f"{table.key} is set as upgraded to {value['value']}")
                if self.exists_in_uc(table, value["value"]):
                    logger.info(
                        f"The table {table.key} was previously migrated to {value['value']}. "
                        f"To revert the table and allow it to be migrated again use the CLI command:"
                        f"databricks labs ucx revert --schema {table.database} --table {table.name}"
                    )
                    return None
                logger.info(f"The upgrade_to target for {table.key} is missing. Unsetting the upgrade_to property")
                try:
                    self._sql_backend.execute(table.sql_unset_upgraded_to())
                except DatabricksError as err:
                    logger.warning(f"Failed to unset upgraded_to property for Table {table.key}: {err}")

        return table_to_migrate

    def _get_table_properties(self, table: Table):
        table_identifier = f"{escape_sql_identifier(table.database)}.{escape_sql_identifier(table.name)}"
        if table.is_table_in_mount:
            table_identifier = f"delta.`{table.location}`"

        try:
            return self._sql_backend.fetch(f"SHOW TBLPROPERTIES {table_identifier}")
        except DatabricksError as err:
            logger.warning(f"Failed to get properties for Table {table.key}: {err}")
            return None

    def exists_in_uc(self, src_table: Table, target_key: str):
        # Attempts to get the target table info from UC returns True if it exists.
        table_info = self._try_get_table_in_uc(target_key)
        if not table_info:
            return False
        # Corner case for tables in mounts where the table exists in UC, but the location is not the same
        # from the one provided in the mapping
        if src_table.is_table_in_mount and table_info.storage_location != src_table.location:
            raise ResourceConflict(
                f"Expected to be migrated from {src_table.key}, but got {table_info.storage_location}. "
                "You can skip this error using the CLI command: "
                "databricks labs ucx skip "
                f"--schema {src_table.database} --table {src_table.name}"
            )
        if not table_info.properties:
            return True
        upgraded_from = table_info.properties.get("upgraded_from")
        if upgraded_from and upgraded_from != src_table.key and not src_table.is_table_in_mount:
            raise ResourceConflict(
                f"Expected to be migrated from {src_table.key}, but got {upgraded_from}. "
                "You can skip this error using the CLI command: "
                "databricks labs ucx skip "
                f"--schema {src_table.database} --table {src_table.name}"
            )
        return True

    def _try_get_table_in_uc(self, target_key: str):
        try:
            return self._ws.tables.get(target_key)
        except (NotFound, InvalidParameterValue):
            return None



import logging
from abc import ABC
from collections.abc import Iterable

from astroid import Call, InferenceError, NodeNG  # type: ignore
from sqlglot.expressions import Alter, Create, Delete, Drop, Expression, Identifier, Insert, Literal, Select

from databricks.labs.ucx.source_code.base import (
    Advice,
    Deprecation,
    CurrentSessionState,
    PythonLinter,
    SqlLinter,
    DfsaPyCollector,
    DirectFsAccessNode,
    DfsaSqlCollector,
    DirectFsAccess,
)
from databricks.labs.ucx.source_code.python.python_ast import Tree, TreeVisitor, TreeHelper
from databricks.labs.ucx.source_code.python.python_infer import InferredValue
from databricks.labs.ucx.source_code.sql.sql_parser import SqlParser, SqlExpression

logger = logging.getLogger(__name__)


class DirectFsAccessPattern(ABC):

    def __init__(self, prefix: str, allowed_roots: list[str]):
        self._prefix = prefix
        self._allowed_roots = allowed_roots

    def matches(self, value: str) -> bool:
        return value.startswith(self._prefix) and not self._matches_allowed_root(value)

    def _matches_allowed_root(self, value: str):
        return any(value.startswith(f"{self._prefix}/{root}") for root in self._allowed_roots)


class RootPattern(DirectFsAccessPattern):

    def _matches_allowed_root(self, value: str):
        return any(value.startswith(f"/{root}") for root in self._allowed_roots)


# the below aims to implement https://docs.databricks.com/en/files/index.html
DIRECT_FS_ACCESS_PATTERNS = [
    DirectFsAccessPattern("dbfs:/", []),
    DirectFsAccessPattern("file:/", ["Workspace/"]),
    DirectFsAccessPattern("s3:/", []),
    DirectFsAccessPattern("s3n:/", []),
    DirectFsAccessPattern("s3a:/", []),
    DirectFsAccessPattern("wasb:/", []),
    DirectFsAccessPattern("wasbs:/", []),
    DirectFsAccessPattern("abfs:/", []),
    DirectFsAccessPattern("abfss:/", []),
    DirectFsAccessPattern("hdfs:/", []),
    # "/mnt/" is detected by the below pattern,
    RootPattern("/", ["Volumes/", "Workspace/"]),
]


class _DetectDirectFsAccessVisitor(TreeVisitor):
    """
    Visitor that detects file system paths in Python code and checks them
    against a list of known deprecated paths.
    """

    def __init__(self, session_state: CurrentSessionState, prevent_spark_duplicates: bool) -> None:
        self._session_state = session_state
        self._directfs_nodes: list[DirectFsAccessNode] = []
        self._prevent_spark_duplicates = prevent_spark_duplicates

    def visit_call(self, node: Call):
        for arg in node.args:
            self._visit_arg(node, arg)

    def _visit_arg(self, call: Call, arg: NodeNG):
        try:
            for inferred in InferredValue.infer_from_node(arg, self._session_state):
                if not inferred.is_inferred():
                    logger.debug(f"Could not infer value of {arg.as_string()}")
                    continue
                self._check_str_arg(call, arg, inferred)
        except InferenceError as e:
            logger.debug(f"Could not infer value of {arg.as_string()}", exc_info=e)

    def _check_str_arg(self, call_node: Call, arg_node: NodeNG, inferred: InferredValue):
        value = inferred.as_string()
        for pattern in DIRECT_FS_ACCESS_PATTERNS:
            if not pattern.matches(value):
                continue
            # only capture 'open' calls or calls originating from spark or dbutils
            # because there is no other known way to manipulate data directly from file system
            tree = Tree(call_node)
            is_open = TreeHelper.get_call_name(call_node) == "open" and tree.is_builtin()
            is_from_db_utils = False if is_open else tree.is_from_module("dbutils")
            is_from_spark = False if is_open or is_from_db_utils else tree.is_from_module("spark")
            if not (is_open or is_from_db_utils or is_from_spark):
                return
            # avoid duplicate advices that are reported by SparkSqlPyLinter
            if self._prevent_spark_duplicates and is_from_spark:
                return
            # since we're normally filtering out spark calls, we're dealing with dfsas we know little about
            # notably we don't know is_read or is_write
            dfsa = DirectFsAccess(
                path=value,
                is_read=True,
                is_write=False,
            )
            self._directfs_nodes.append(DirectFsAccessNode(dfsa, arg_node))
            return

    @property
    def directfs_nodes(self):
        return self._directfs_nodes


class DirectFsAccessPyLinter(PythonLinter, DfsaPyCollector):

    def __init__(self, session_state: CurrentSessionState, prevent_spark_duplicates=True):
        self._session_state = session_state
        self._prevent_spark_duplicates = prevent_spark_duplicates

    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
        """
        Lints the code looking for file system paths that are deprecated
        """
        visitor = _DetectDirectFsAccessVisitor(self._session_state, self._prevent_spark_duplicates)
        visitor.visit(tree.node)
        for directfs_node in visitor.directfs_nodes:
            advisory = Deprecation.from_node(
                code='direct-filesystem-access',
                message=f"The use of direct filesystem references is deprecated: {directfs_node.dfsa.path}",
                node=directfs_node.node,
            )
            yield advisory

    def collect_dfsas_from_tree(self, tree: Tree) -> Iterable[DirectFsAccessNode]:
        visitor = _DetectDirectFsAccessVisitor(self._session_state, self._prevent_spark_duplicates)
        visitor.visit(tree.node)
        yield from visitor.directfs_nodes


class DirectFsAccessSqlLinter(SqlLinter, DfsaSqlCollector):

    def lint_expression(self, expression: Expression):
        for dfsa in self._collect_dfsas(SqlExpression(expression)):
            yield Deprecation(
                code='direct-filesystem-access-in-sql-query',
                message=f"The use of direct filesystem references is deprecated: {dfsa.path}",
                # SQLGlot does not propagate tokens yet. See https://github.com/tobymao/sqlglot/issues/3159
                start_line=0,
                start_col=0,
                end_line=0,
                end_col=1024,
            )

    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]:
        yield from SqlParser.walk_expressions(source_code, self._collect_dfsas)

    @classmethod
    def _collect_dfsas(cls, expression: SqlExpression) -> Iterable[DirectFsAccess]:
        yield from cls._collect_dfsas_from_literals(expression)
        yield from cls._collect_dfsas_from_identifiers(expression)

    @classmethod
    def _collect_dfsas_from_literals(cls, expression: SqlExpression) -> Iterable[DirectFsAccess]:
        for literal in expression.find_all(Literal):
            if not isinstance(literal.this, str):
                logger.warning(f"Can't interpret {type(literal.this).__name__}")
            yield from cls._collect_dfsa_from_node(literal, literal.this)

    @classmethod
    def _collect_dfsas_from_identifiers(cls, expression: SqlExpression) -> Iterable[DirectFsAccess]:
        for identifier in expression.find_all(Identifier):
            if not isinstance(identifier.this, str):
                logger.warning(f"Can't interpret {type(identifier.this).__name__}")
            yield from cls._collect_dfsa_from_node(identifier, identifier.this)

    @classmethod
    def _collect_dfsa_from_node(cls, expression: Expression, path: str) -> Iterable[DirectFsAccess]:
        if any(pattern.matches(path) for pattern in DIRECT_FS_ACCESS_PATTERNS):
            is_read = cls._is_read(expression)
            is_write = cls._is_write(expression)
            yield DirectFsAccess(
                path=path,
                is_read=is_read,
                is_write=is_write,
            )

    @classmethod
    def _is_read(cls, expression: Expression | None) -> bool:
        expression = cls._walk_up(expression)
        return isinstance(expression, Select)

    @classmethod
    def _is_write(cls, expression: Expression | None) -> bool:
        expression = cls._walk_up(expression)
        return isinstance(expression, (Create, Alter, Drop, Insert, Delete))

    @classmethod
    def _walk_up(cls, expression: Expression | None) -> Expression | None:
        if expression is None:
            return None
        if isinstance(expression, (Create, Alter, Drop, Insert, Delete, Select)):
            return expression
        return cls._walk_up(expression.parent)



from collections.abc import Iterator

from databricks.labs.lsql.backends import SqlBackend
from databricks.labs.lsql.core import Row

from .base import DataProfiler, DataProfilingResult, TableIdentifier, TableMetadataRetriever


class StandardDataProfiler(DataProfiler):
    def __init__(self, sql_backend: SqlBackend, metadata_retriever: TableMetadataRetriever):
        self._sql_backend = sql_backend
        self._metadata_retriever = metadata_retriever

    def profile_data(self, entity: TableIdentifier) -> DataProfilingResult:
        """
        This method profiles the data in the given table. It takes a TableIdentifier object as input, which represents
        the table to be profiled. The method performs two main operations:

        1. It retrieves the row count of the table.
        2. It retrieves the metadata of the table using a TableMetadataRetriever instance.

        Note: This method does not handle exceptions raised during the execution of the SQL query or the retrieval
        of the table metadata. These exceptions are expected to be handled by the caller
        in a manner appropriate for their context.
        """
        row_count = self._get_table_row_count(entity)
        return DataProfilingResult(
            row_count,
            self._metadata_retriever.get_metadata(entity),
        )

    def _get_table_row_count(self, entity: TableIdentifier) -> int:
        query_result: Iterator[Row] = self._sql_backend.fetch(f"SELECT COUNT(*) as row_count FROM {entity.fqn_escaped}")
        count_row = next(query_result)
        return int(count_row[0])



import logging

from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import PermissionDenied

logger = logging.getLogger(__name__)


class VerifyHasMetastore:
    def __init__(self, ws: WorkspaceClient):
        self.metastore_id: str | None = None
        self.default_catalog_name: str | None = None
        self.workspace_id: int | None = None
        self._ws = ws

    def verify_metastore(self):
        """
        Verifies if a metastore exists for a metastore
        :param :
        :return:
        """

        try:
            current_metastore = self._ws.metastores.current()
            if current_metastore:
                self.default_catalog_name = current_metastore.default_catalog_name
                self.metastore_id = current_metastore.metastore_id
                self.workspace_id = current_metastore.workspace_id
                return True
            raise MetastoreNotFoundError
        except PermissionDenied as err:
            logger.error(f"Cannot access metastore: {err}")
            return False


class MetastoreNotFoundError(Exception):
    def __init__(self, message="Metastore not found in the workspace"):
        self.message = message
        super().__init__(self.message)



import contextlib
import datetime as dt
import logging
import os
import re
from collections.abc import Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import timedelta
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from typing import TextIO

from databricks.labs.blueprint.logger import install_logger
from databricks.labs.lsql.backends import SqlBackend

from databricks.sdk.errors import InternalError
from databricks.sdk.retries import retried

from databricks.labs.ucx.__about__ import __version__

logger = logging.getLogger(__name__)


@dataclass
class LogRecord:
    timestamp: int
    job_id: int
    workflow_name: str
    task_name: str
    job_run_id: int
    level: str
    component: str
    message: str


@dataclass
class PartialLogRecord:
    """The information found within a log file record."""

    time: dt.time
    level: str
    component: str
    message: str


def peak_multi_line_message(log: TextIO, pattern: re.Pattern) -> tuple[str, re.Match | None, str]:
    """
    A single log record message may span multiple log lines. In this case, the regex on
    subsequent lines do not match.

    Args:
         log (TextIO): The log file IO.
         pattern (re.Pattern): The regex pattern for a log line.
    """
    multi_line_message = ""
    line = log.readline()
    match = pattern.match(line)
    while len(line) > 0 and match is None:
        multi_line_message += "\n" + line.rstrip()
        line = log.readline()
        match = pattern.match(line)
    return line, match, multi_line_message


def parse_logs(log: TextIO) -> Iterator[PartialLogRecord]:
    """Parse the logs to retrieve values for PartialLogRecord fields.

    Args:
         log (TextIO): The log file IO.
    """
    time_format = "%H:%M:%S"
    # This regex matches the log format defined in databricks.labs.ucx.installer.logs.TaskLogger
    log_format = r"(\d+:\d+:\d+)\s(\w+)\s\[(.+)\]\s\{\w+\}\s(.+)"
    pattern = re.compile(log_format)

    line = log.readline()
    match = pattern.match(line)
    if match is None:
        logger.warning(f"Logs do not match expected format ({log_format}): {line}")
        return
    while len(line) > 0:
        assert match is not None
        time, *groups, message = match.groups()

        next_line, next_match, multi_line_message = peak_multi_line_message(log, pattern)

        time = dt.datetime.strptime(time, time_format).time()
        # Mypy can't determine length of regex expressions
        partial_log_record = PartialLogRecord(time, *groups, message + multi_line_message)  # type: ignore

        yield partial_log_record

        line, match = next_line, next_match


class TaskRunWarningRecorder:
    def __init__(
        self,
        install_dir: Path | str,
        workflow: str,
        job_id: int,
        job_run_id: int,
        sql_backend: SqlBackend,
        schema: str,
        attempt: int = 0,
    ):
        """
        Initializes a LogProcessor instance.

        Args:
            install_dir (Path | str): The installation folder on WorkspaceFS
            workflow (str): The workflow name.
            job_id (int): The job id of the job to store the log records for.
            job_run_id (int): The job run id of the job to store the log records for.
            sql_backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
            schema (str): The schema name for the logs persistence.
        """
        self._catalog = "hive_metastore"
        self._table = "logs"

        self._workflow = workflow
        self._job_id = job_id
        self._job_run_id = job_run_id
        self._sql_backend = sql_backend
        self._schema = schema

        self._log_path = TaskLogger.log_path(Path(str(install_dir)), workflow, job_run_id, attempt)

    @property
    def full_name(self) -> str:
        return f"{self._catalog}.{self._schema}.{self._table}"

    def _record_task(self, task_name: str, log: TextIO, log_creation_timestamp: dt.datetime) -> list[LogRecord]:
        """Record the logs of a given task.

        Args:
            task_name (str): The name of the task
            log (TextIO): The task log
            log_creation_timestamp (datetime.datetime): The log creation timestamp.
        """
        log_records = []
        for partial_log_record in parse_logs(log):
            if logging.getLevelName(partial_log_record.level) < logging.WARNING:
                continue
            log_record = LogRecord(
                timestamp=int(
                    log_creation_timestamp.replace(
                        hour=partial_log_record.time.hour,
                        minute=partial_log_record.time.minute,
                        second=partial_log_record.time.second,
                    ).timestamp()
                ),
                job_id=self._job_id,
                workflow_name=self._workflow,
                task_name=task_name,
                job_run_id=self._job_run_id,
                level=partial_log_record.level,
                component=partial_log_record.component,
                message=partial_log_record.message,
            )
            log_records.append(log_record)
        return log_records

    def snapshot(self) -> list[LogRecord]:
        """Parse and store the logs for all tasks.

        Raises:
            InternalError: When error or above log records are found, after storing the logs.
        """
        log_files = self._log_path.glob("*.log*")

        log_records = []
        for log_file in log_files:
            task_name = log_file.stem.rstrip(".log")
            logger.debug(f"Parsing logs for {task_name} in file {log_file}")

            log_creation_timestamp = dt.datetime.utcfromtimestamp(log_file.stat().st_ctime)
            with log_file.open("r") as log:
                log_records += self._record_task(task_name, log, log_creation_timestamp)

        self._sql_backend.save_table(
            self.full_name,
            log_records,
            LogRecord,
            mode="append",
        )

        error_messages = []
        for log_record in log_records:
            if logging.getLevelName(log_record.level) < logging.ERROR:
                continue
            message = log_record.message
            # Ignore the prompt to troubleshoot, which is the error
            if 'databricks workspace export /' in message:
                continue
            # This error comes up during testing, when ucx crawls tables & schemas created by other tests,
            # but couldn't fetch the grants later as they have already been dropped by those tests. Ignore them
            errors_in_test = ["Couldn't fetch grants"]
            if any(error in message for error in errors_in_test) and self._is_testing():
                continue
            error_messages.append(message)
        if len(error_messages) > 0:
            raise InternalError("\n".join(error_messages))
        return log_records

    @staticmethod
    def _is_testing():
        return "+" in __version__


class TaskLogger(contextlib.AbstractContextManager):
    # files are available in the workspace only once their handlers are closed,
    # so we rotate files log every minute to make them available for download.
    #
    # See https://docs.python.org/3/library/logging.handlers.html#logging.handlers.TimedRotatingFileHandler
    # See https://docs.python.org/3/howto/logging-cookbook.html

    def __init__(
        self,
        install_dir: Path,
        workflow: str,
        workflow_id: str,
        task_name: str,
        workflow_run_id: str,
        log_level="INFO",
        attempt: str = "0",
    ):
        self._log_level = log_level
        self._workflow = workflow
        self._workflow_id = workflow_id
        self._workflow_run_id = workflow_run_id
        self._databricks_logger = logging.getLogger("databricks")
        self._app_logger = logging.getLogger("databricks.labs.ucx")
        self._log_path = self.log_path(install_dir, workflow, workflow_run_id, attempt)
        self.log_file = self._log_path / f"{task_name}.log"
        self._app_logger.info(f"UCX v{__version__} After job finishes, see debug logs at {self.log_file}")

    @classmethod
    def log_path(cls, install_dir: Path, workflow: str, workflow_run_id: str | int, attempt: str | int) -> Path:
        return install_dir / "logs" / workflow / f"run-{workflow_run_id}-{attempt}"

    def __repr__(self):
        return self.log_file.as_posix()

    def __enter__(self):
        self._log_path.mkdir(parents=True, exist_ok=True)
        self._init_debug_logfile()
        self._init_run_readme()
        self._databricks_logger.setLevel(logging.DEBUG)
        self._app_logger.setLevel(logging.DEBUG)
        console_handler = install_logger(self._log_level)
        self._databricks_logger.removeHandler(console_handler)
        self._databricks_logger.addHandler(self._file_handler)
        return self

    def __exit__(self, _t, error, _tb):
        if error:
            log_file_for_cli = str(self.log_file).removeprefix("/Workspace")
            cli_command = f"databricks workspace export /{log_file_for_cli}"
            self._app_logger.error(f"Execute `{cli_command}` locally to troubleshoot with more details. {error}")
            self._databricks_logger.debug("Task crash details", exc_info=error)
        self._file_handler.flush()
        self._file_handler.close()

    def _init_debug_logfile(self):
        log_format = "%(asctime)s %(levelname)s [%(name)s] {%(threadName)s} %(message)s"
        log_formatter = logging.Formatter(fmt=log_format, datefmt="%H:%M:%S")
        self._file_handler = TimedRotatingFileHandler(self.log_file.as_posix(), when="M", interval=1)
        self._file_handler.setFormatter(log_formatter)
        self._file_handler.setLevel(logging.DEBUG)

    def _init_run_readme(self):
        log_readme = self._log_path.joinpath("README.md")
        if log_readme.exists():
            return
        # this may race when run from multiple tasks, therefore it must be multiprocess safe
        with self._exclusive_open(str(log_readme), mode="w") as f:
            f.write(f"# Logs for the UCX {self._workflow} workflow\n")
            f.write("This folder contains UCX log files.\n\n")
            f.write(f"See the [{self._workflow} job](/#job/{self._workflow_id}) and ")
            f.write(f"[run #{self._workflow_run_id}](/#job/{self._workflow_id}/run/{self._workflow_run_id})\n")

    @classmethod
    @contextmanager
    def _exclusive_open(cls, filename: str, **kwargs):
        """Open a file with exclusive access across multiple processes.
        Requires write access to the directory containing the file.

        Arguments are the same as the built-in open.

        Returns a context manager that closes the file and releases the lock.
        """
        lockfile_name = filename + ".lock"
        lockfile = cls._create_lock(lockfile_name)

        try:
            with open(filename, encoding="utf8", **kwargs) as f:
                yield f
        finally:
            try:
                os.close(lockfile)
            finally:
                os.unlink(lockfile_name)

    @staticmethod
    @retried(on=[FileExistsError], timeout=timedelta(seconds=5))
    def _create_lock(lockfile_name):
        while True:  # wait until the lock file can be opened
            f = os.open(lockfile_name, os.O_CREAT | os.O_EXCL)
            break
        return f



import dataclasses
import logging
from collections.abc import Iterable
from dataclasses import asdict, dataclass, field
from datetime import datetime, timezone

from databricks.sdk import WorkspaceClient
from databricks.sdk.service.sql import Dashboard, LegacyQuery
from databricks.sdk.service.workspace import Language

from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex
from databricks.labs.ucx.source_code.base import CurrentSessionState, LineageAtom, UsedTable
from databricks.labs.ucx.source_code.directfs_access import DirectFsAccessCrawler, DirectFsAccess
from databricks.labs.ucx.source_code.linters.context import LinterContext
from databricks.labs.ucx.source_code.redash import Redash
from databricks.labs.ucx.source_code.used_table import UsedTablesCrawler

logger = logging.getLogger(__name__)


@dataclass
class QueryProblem:
    dashboard_id: str
    dashboard_parent: str
    dashboard_name: str
    query_id: str
    query_parent: str
    query_name: str
    code: str
    message: str


@dataclass
class _ReportingContext:
    linted_queries: set[str] = field(default_factory=set)
    all_problems: list[QueryProblem] = field(default_factory=list)
    all_dfsas: list[DirectFsAccess] = field(default_factory=list)
    all_tables: list[UsedTable] = field(default_factory=list)


class QueryLinter:

    def __init__(
        self,
        ws: WorkspaceClient,
        migration_index: TableMigrationIndex,
        directfs_crawler: DirectFsAccessCrawler,
        used_tables_crawler: UsedTablesCrawler,
        include_dashboard_ids: list[str] | None,
    ):
        self._ws = ws
        self._migration_index = migration_index
        self._directfs_crawler = directfs_crawler
        self._used_tables_crawler = used_tables_crawler
        self._include_dashboard_ids = include_dashboard_ids

    def refresh_report(self, sql_backend: SqlBackend, inventory_database: str):
        assessment_start = datetime.now(timezone.utc)
        context = _ReportingContext()
        self._lint_dashboards(context)
        self._lint_queries(context)
        assessment_end = datetime.now(timezone.utc)
        self._dump_problems(context, sql_backend, inventory_database)
        self._dump_dfsas(context, assessment_start, assessment_end)
        self._dump_used_tables(context, assessment_start, assessment_end)

    def _dump_problems(self, context: _ReportingContext, sql_backend: SqlBackend, inventory_database: str):
        logger.info(f"Saving {len(context.all_problems)} linting problems...")
        sql_backend.save_table(
            f'{escape_sql_identifier(inventory_database)}.query_problems',
            context.all_problems,
            QueryProblem,
            mode='overwrite',
        )

    def _dump_dfsas(self, context: _ReportingContext, assessment_start: datetime, assessment_end: datetime):
        processed_dfsas = []
        for dfsa in context.all_dfsas:
            dfsa = dataclasses.replace(
                dfsa,
                assessment_start_timestamp=assessment_start,
                assessment_end_timestamp=assessment_end,
            )
            processed_dfsas.append(dfsa)
        self._directfs_crawler.dump_all(processed_dfsas)

    def _dump_used_tables(self, context: _ReportingContext, assessment_start: datetime, assessment_end: datetime):
        processed_tables = []
        for table in context.all_tables:
            table = dataclasses.replace(
                table,
                assessment_start_timestamp=assessment_start,
                assessment_end_timestamp=assessment_end,
            )
            processed_tables.append(table)
        self._used_tables_crawler.dump_all(processed_tables)

    def _lint_dashboards(self, context: _ReportingContext):
        dashboard_ids = self._dashboard_ids_in_scope()
        logger.info(f"Running {len(dashboard_ids)} linting tasks...")
        for dashboard_id in dashboard_ids:
            dashboard = self._ws.dashboards.get(dashboard_id=dashboard_id)
            problems, dfsas, tables = self._lint_and_collect_from_dashboard(dashboard, context.linted_queries)
            context.all_problems.extend(problems)
            context.all_dfsas.extend(dfsas)
            context.all_tables.extend(tables)

    def _lint_queries(self, context: _ReportingContext):
        for query in self._queries_in_scope():
            if query.id in context.linted_queries:
                continue
            context.linted_queries.add(query.id)
            problems = self.lint_query(query)
            context.all_problems.extend(problems)
            dfsas = self.collect_dfsas_from_query("no-dashboard-id", query)
            context.all_dfsas.extend(dfsas)
            tables = self.collect_used_tables_from_query("no-dashboard-id", query)
            context.all_tables.extend(tables)

    def _dashboard_ids_in_scope(self) -> list[str]:
        if self._include_dashboard_ids is not None:  # an empty list is accepted
            return self._include_dashboard_ids
        all_dashboards = self._ws.dashboards.list()
        return [dashboard.id for dashboard in all_dashboards if dashboard.id]

    def _queries_in_scope(self):
        if self._include_dashboard_ids is not None:  # an empty list is accepted
            return []
        all_queries = self._ws.queries_legacy.list()
        return [query for query in all_queries if query.id]

    def _lint_and_collect_from_dashboard(
        self, dashboard: Dashboard, linted_queries: set[str]
    ) -> tuple[Iterable[QueryProblem], Iterable[DirectFsAccess], Iterable[UsedTable]]:
        dashboard_queries = Redash.get_queries_from_dashboard(dashboard)
        query_problems: list[QueryProblem] = []
        query_dfsas: list[DirectFsAccess] = []
        query_tables: list[UsedTable] = []
        dashboard_id = dashboard.id or "<no-id>"
        dashboard_parent = dashboard.parent or "<orphan>"
        dashboard_name = dashboard.name or "<anonymous>"
        for query in dashboard_queries:
            if query.id is None:
                continue
            if query.id in linted_queries:
                continue
            linted_queries.add(query.id)
            problems = self.lint_query(query)
            for problem in problems:
                query_problems.append(
                    dataclasses.replace(
                        problem,
                        dashboard_id=dashboard_id,
                        dashboard_parent=dashboard_parent,
                        dashboard_name=dashboard_name,
                    )
                )
            dfsas = self.collect_dfsas_from_query(dashboard_id, query)
            for dfsa in dfsas:
                atom = LineageAtom(
                    object_type="DASHBOARD",
                    object_id=dashboard_id,
                    other={"parent": dashboard_parent, "name": dashboard_name},
                )
                source_lineage = [atom] + dfsa.source_lineage
                query_dfsas.append(dataclasses.replace(dfsa, source_lineage=source_lineage))
            tables = self.collect_used_tables_from_query(dashboard_id, query)
            for table in tables:
                atom = LineageAtom(
                    object_type="DASHBOARD",
                    object_id=dashboard_id,
                    other={"parent": dashboard_parent, "name": dashboard_name},
                )
                source_lineage = [atom] + table.source_lineage
                query_tables.append(dataclasses.replace(table, source_lineage=source_lineage))
        return query_problems, query_dfsas, query_tables

    def lint_query(self, query: LegacyQuery) -> Iterable[QueryProblem]:
        if not query.query:
            return
        ctx = LinterContext(self._migration_index, CurrentSessionState())
        linter = ctx.linter(Language.SQL)
        query_id = query.id or "<no-id>"
        query_parent = query.parent or "<orphan>"
        query_name = query.name or "<anonymous>"
        for advice in linter.lint(query.query):
            yield QueryProblem(
                dashboard_id="",
                dashboard_parent="",
                dashboard_name="",
                query_id=query_id,
                query_parent=query_parent,
                query_name=query_name,
                code=advice.code,
                message=advice.message,
            )

    def collect_dfsas_from_query(self, dashboard_id: str, query: LegacyQuery) -> Iterable[DirectFsAccess]:
        if query.query is None:
            return
        ctx = LinterContext(self._migration_index, CurrentSessionState())
        collector = ctx.dfsa_collector(Language.SQL)
        source_id = f"{dashboard_id}/{query.id}"
        source_name = query.name or "<anonymous>"
        source_timestamp = self._read_timestamp(query.updated_at)
        source_lineage = [LineageAtom(object_type="QUERY", object_id=source_id, other={"name": source_name})]
        for dfsa in collector.collect_dfsas(query.query):
            yield DirectFsAccess(**asdict(dfsa)).replace_source(
                source_id=source_id, source_timestamp=source_timestamp, source_lineage=source_lineage
            )

    def collect_used_tables_from_query(self, dashboard_id: str, query: LegacyQuery) -> Iterable[UsedTable]:
        if query.query is None:
            return
        ctx = LinterContext(self._migration_index, CurrentSessionState())
        collector = ctx.tables_collector(Language.SQL)
        source_id = f"{dashboard_id}/{query.id}"
        source_name = query.name or "<anonymous>"
        source_timestamp = self._read_timestamp(query.updated_at)
        source_lineage = [LineageAtom(object_type="QUERY", object_id=source_id, other={"name": source_name})]
        for table in collector.collect_tables(query.query):
            yield UsedTable(**asdict(table)).replace_source(
                source_id=source_id, source_timestamp=source_timestamp, source_lineage=source_lineage
            )

    @classmethod
    def _read_timestamp(cls, timestamp: str | None) -> datetime:
        if timestamp is not None:
            methods = [
                datetime.fromisoformat,
                lambda s: datetime.fromisoformat(s[:-1]),  # ipython breaks on final 'Z'
            ]
            for method in methods:
                try:
                    return method(timestamp)
                except ValueError:
                    pass
        return datetime.now()



from __future__ import annotations

import abc
import dataclasses
import itertools
import logging
from dataclasses import dataclass
from pathlib import Path
from collections.abc import Callable, Iterable, Iterator
from typing import TypeVar, Generic

from astroid import (  # type: ignore
    NodeNG,
)
from databricks.labs.ucx.source_code.base import Advisory, CurrentSessionState, is_a_notebook, LineageAtom
from databricks.labs.ucx.source_code.python.python_ast import Tree
from databricks.labs.ucx.source_code.path_lookup import PathLookup

logger = logging.Logger(__name__)


class DependencyGraph:

    def __init__(
        self,
        dependency: Dependency,
        parent: DependencyGraph | None,
        resolver: DependencyResolver,
        path_lookup: PathLookup,
        session_state: CurrentSessionState,
    ):
        self._dependency = dependency
        self._parent = parent
        self._resolver = resolver
        self._path_lookup = path_lookup.change_directory(dependency.path.parent)
        self._session_state = session_state
        self._dependencies: dict[Dependency, DependencyGraph] = {}

    @property
    def path_lookup(self):
        return self._path_lookup

    @property
    def dependency(self):
        return self._dependency

    @property
    def dependencies(self):
        return self._dependencies

    def register_library(self, *libraries: str) -> list[DependencyProblem]:
        return self._resolver.register_library(self.path_lookup, *libraries)

    def register_notebook(self, path: Path, inherit_context: bool) -> list[DependencyProblem]:
        maybe = self._resolver.resolve_notebook(self.path_lookup, path, inherit_context)
        if not maybe.dependency:
            return maybe.problems
        maybe_graph = self.register_dependency(maybe.dependency)
        return maybe_graph.problems

    def register_import(self, name: str) -> list[DependencyProblem]:
        if not name:
            return [DependencyProblem('import-empty', 'Empty import name')]
        maybe = self._resolver.resolve_import(self.path_lookup, name)
        if not maybe.dependency:
            return maybe.problems
        maybe_graph = self.register_dependency(maybe.dependency)
        return maybe_graph.problems

    def register_file(self, path: Path) -> list[DependencyProblem]:
        maybe = self._resolver.resolve_file(self.path_lookup, path)
        if not maybe.dependency:
            return maybe.problems
        maybe_graph = self.register_dependency(maybe.dependency)
        return maybe_graph.problems

    def register_dependency(self, dependency: Dependency) -> MaybeGraph:
        # TODO: this has to be a private method, because we don't want to allow free-form dependencies.
        # the only case we have for this method to be used outside of this class is for DistInfoPackage
        # See databricks.labs.ucx.source_code.python_libraries.DistInfoContainer.build_dependency_graph for reference
        logger.debug(f"Registering dependency {dependency}")
        maybe = self.locate_dependency(dependency.path)
        if maybe.graph is not None:
            self._dependencies[dependency] = maybe.graph
            return maybe
        # nay, create the child graph and populate it
        child_graph = DependencyGraph(dependency, self, self._resolver, self._path_lookup, self._session_state)
        self._dependencies[dependency] = child_graph
        container = dependency.load(self.path_lookup)
        # TODO: Return either (child) graph OR problems
        if not container:
            problem = DependencyProblem('dependency-register-failed', 'Failed to register dependency', dependency.path)
            return MaybeGraph(child_graph, [problem])
        problems = []
        for problem in container.build_dependency_graph(child_graph):
            if problem.is_path_missing():
                problem = dataclasses.replace(problem, source_path=dependency.path)
            problems.append(problem)
        return MaybeGraph(child_graph, problems)

    def locate_dependency(self, path: Path) -> MaybeGraph:
        # need a list since unlike JS, Python won't let you assign closure variables
        found: list[DependencyGraph] = []

        def check_registered_dependency(graph):
            if graph.dependency.path == path:
                found.append(graph)
                return True
            return False

        self.root.visit(check_registered_dependency, set())
        if not found:
            return MaybeGraph(None, [DependencyProblem('dependency-not-found', 'Dependency not found')])
        return MaybeGraph(found[0], [])

    @property
    def root(self):
        return self if self._parent is None else self._parent.root

    @property
    def parent(self):
        return self._parent

    @property
    def all_dependencies(self) -> set[Dependency]:
        dependencies: set[Dependency] = set()

        def add_to_dependencies(graph: DependencyGraph) -> bool:
            if graph.dependency in dependencies:
                return True
            dependencies.add(graph.dependency)
            return False

        self.visit(add_to_dependencies, set())
        return dependencies

    @property
    def root_dependencies(self) -> set[Dependency]:
        roots: set[Dependency] = set()
        children: set[Dependency] = set()

        def add_to_dependencies(graph: DependencyGraph) -> bool:
            dependency = graph.dependency
            # if already encountered we're done
            if dependency in children:
                return False
            # if it's not a real file, then it's not a root
            if not dependency.path.is_file() and not is_a_notebook(dependency.path):
                children.add(dependency)
                return False
            # if it appears more than once then it can't be a root
            if dependency in roots:
                roots.remove(dependency)
                children.add(dependency)
                return False
            # if it has a 'real' parent, it's a child
            parent_graph = graph.parent
            while parent_graph is not None:
                dep = parent_graph.dependency
                if dep.path.is_file() or is_a_notebook(dep.path):
                    children.add(dependency)
                    return False
                parent_graph = parent_graph.parent
            # ok, it's a root
            roots.add(dependency)
            return False

        self.visit(add_to_dependencies, None)
        return roots

    @property
    def local_dependencies(self) -> set[Dependency]:
        return {child.dependency for child in self._dependencies.values()}

    def all_relative_names(self) -> set[str]:
        """This method is intended to simplify testing"""
        return self._relative_names(self.all_dependencies)

    def _relative_names(self, dependencies: set[Dependency]) -> set[str]:
        """This method is intended to simplify testing"""
        names = set[str]()
        for dependency in dependencies:
            for library_root in self._path_lookup.library_roots:
                if not dependency.path.is_relative_to(library_root):
                    continue
                relative_path = dependency.path.relative_to(library_root).as_posix()
                if relative_path == ".":
                    continue
                names.add(relative_path)
        return names

    def root_paths(self) -> set[Path]:
        return {d.path for d in self.root_dependencies}

    def root_relative_names(self) -> set[str]:
        """This method is intended to simplify testing"""
        return self._relative_names(self.root_dependencies)

    def visit(self, visit_node: Callable[[DependencyGraph], bool | None], visited: set[Path] | None) -> bool:
        """ "
        when visit_node returns True it interrupts the visit
        provide visited set if you want to ensure nodes are only visited once
        """
        visitor = DependencyGraphVisitor(visit_node, visited)
        return visitor.visit(self)

    def new_dependency_graph_context(self):
        return DependencyGraphContext(
            parent=self, path_lookup=self._path_lookup, resolver=self._resolver, session_state=self._session_state
        )

    def _compute_route(self, root: Path, leaf: Path, visited: set[Path]) -> list[Dependency]:
        """given 2 files or notebooks root and leaf, compute the list of dependencies that must be traversed
        in order to lint the leaf in the context of its parents"""
        try:
            route = self._do_compute_route(root, leaf, visited)
            return self._trim_route(route)
        except ValueError:
            # we only compute routes on graphs so _compute_route can't fail
            # but we return an empty route in case it does
            return []

    def _do_compute_route(self, root: Path, leaf: Path, visited: set[Path]) -> list[Dependency]:
        maybe = self.locate_dependency(root)
        if not maybe.graph:
            logger.warning(f"Could not compute route because dependency {root} cannot be located")
            raise ValueError()
        route: list[Dependency] = []

        def do_compute_route(graph: DependencyGraph) -> bool:
            route.append(graph.dependency)
            for dependency in graph.local_dependencies:
                if dependency.path == leaf:
                    route.append(dependency)
                    return True
            for dependency in graph.local_dependencies:
                sub_route = self._do_compute_route(dependency.path, leaf, visited)
                if len(sub_route) > 0:
                    route.extend(sub_route)
                    return True
            route.pop()
            return False

        maybe.graph.visit(do_compute_route, visited)
        return route

    def _trim_route(self, dependencies: list[Dependency]) -> list[Dependency]:
        """don't inherit context if dependency is not a file/notebook, or it is loaded via dbutils.notebook.run or via import"""
        # filter out intermediate containers
        deps_with_source: list[Dependency] = []
        for dependency in dependencies:
            if dependency.path.is_file() or is_a_notebook(dependency.path):
                deps_with_source.append(dependency)
        # restart when not inheriting context
        for i, dependency in enumerate(deps_with_source):
            if dependency.inherits_context:
                continue
            return [dependency] + self._trim_route(deps_with_source[i + 1 :])
        return deps_with_source

    def build_inherited_tree(self, root: Path, leaf: Path) -> Tree | None:
        return self._build_inherited_context(root, leaf).tree

    def _build_inherited_context(self, root: Path, leaf: Path) -> InheritedContext:
        route = self._compute_route(root, leaf, set())
        return InheritedContext.from_route(self, self.path_lookup, route)

    def __repr__(self):
        return f"<DependencyGraph {self.dependency.path}>"


class DependencyGraphVisitor:

    def __init__(self, visit_node: Callable[[DependencyGraph], bool | None], visited: set[Path] | None):
        self._visit_node = visit_node
        self._visited = visited
        self._visited_pairs: set[tuple[Path, Path]] = set()

    def visit(self, graph: DependencyGraph) -> bool:
        path = graph.dependency.path
        if self._visited is not None:
            if path in self._visited:
                return False
            self._visited.add(path)
        if self._visit_node(graph):
            return True
        for dependency_graph in graph.dependencies.values():
            pair = (path, dependency_graph.dependency.path)
            if pair in self._visited_pairs:
                continue
            self._visited_pairs.add(pair)
            if self.visit(dependency_graph):
                return True
        return False


@dataclass
class DependencyGraphContext:
    parent: DependencyGraph
    path_lookup: PathLookup
    resolver: DependencyResolver
    session_state: CurrentSessionState


class Dependency:

    def __init__(self, loader: DependencyLoader, path: Path, inherits_context=True):
        self._loader = loader
        self._path = path
        self._inherits_context = inherits_context

    @property
    def path(self) -> Path:
        return self._path

    @property
    def inherits_context(self):
        return self._inherits_context

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.path == other.path

    def load(self, path_lookup: PathLookup) -> SourceContainer | None:
        return self._loader.load_dependency(path_lookup, self)

    def __repr__(self):
        return f"Dependency<{self.path}>"

    @property
    def lineage(self) -> list[LineageAtom]:
        object_type = "NOTEBOOK" if is_a_notebook(self.path) else "FILE"
        return [LineageAtom(object_type=object_type, object_id=str(self.path))]


class SourceContainer(abc.ABC):

    @abc.abstractmethod
    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]: ...

    def build_inherited_context(self, graph: DependencyGraph, child_path: Path) -> InheritedContext:
        raise ValueError(f"Building an inherited context from {type(self).__name__} is not supported!")


class DependencyLoader(abc.ABC):
    @abc.abstractmethod
    def load_dependency(self, path_lookup: PathLookup, dependency: Dependency) -> SourceContainer | None: ...


class WrappingLoader(DependencyLoader):

    def __init__(self, source_container: SourceContainer):
        self._source_container = source_container

    def load_dependency(self, path_lookup: PathLookup, dependency: Dependency) -> SourceContainer | None:
        return self._source_container

    def __repr__(self):
        return f"<WrappingLoader source_container={self._source_container}>"


class LibraryResolver(abc.ABC):
    @abc.abstractmethod
    def register_library(self, path_lookup: PathLookup, *libraries: str) -> list[DependencyProblem]:
        pass


class BaseNotebookResolver(abc.ABC):

    @abc.abstractmethod
    def resolve_notebook(self, path_lookup: PathLookup, path: Path, inherit_context: bool) -> MaybeDependency: ...

    @staticmethod
    def _fail(code: str, message: str) -> MaybeDependency:
        return MaybeDependency(None, [DependencyProblem(code, message)])


class BaseImportResolver(abc.ABC):

    @abc.abstractmethod
    def resolve_import(self, path_lookup: PathLookup, name: str) -> MaybeDependency:
        """resolve a simple or composite import name"""


class BaseFileResolver(abc.ABC):

    @abc.abstractmethod
    def resolve_file(self, path_lookup, path: Path) -> MaybeDependency:
        """locates a file"""


@dataclass
class MaybeDependency:
    dependency: Dependency | None
    problems: list[DependencyProblem]


class DependencyResolver:
    """the DependencyResolver is responsible for locating "stuff", mimicking the Databricks runtime behavior.
    There are specific resolution algorithms for import and for Notebooks (executed via %run or dbutils.notebook.run)
    so we're using 2 distinct resolvers for notebooks (instance) and imports (linked list of specialized sub-resolvers)
    resolving imports is affected by cwd and sys.paths, the latter being itself influenced by Python code
    we therefore need a PathLookup to convey these during import resolution
    """

    def __init__(
        self,
        library_resolver: LibraryResolver,
        notebook_resolver: BaseNotebookResolver,
        import_resolver: BaseImportResolver,
        file_resolver: BaseFileResolver,
        path_lookup: PathLookup,
    ):
        self._library_resolver = library_resolver
        self._notebook_resolver = notebook_resolver
        self._import_resolver = import_resolver
        self._file_resolver = file_resolver
        self._path_lookup = path_lookup

    def resolve_notebook(self, path_lookup: PathLookup, path: Path, inherit_context: bool) -> MaybeDependency:
        return self._notebook_resolver.resolve_notebook(path_lookup, path, inherit_context)

    def resolve_import(self, path_lookup: PathLookup, name: str) -> MaybeDependency:
        return self._import_resolver.resolve_import(path_lookup, name)

    def resolve_file(self, path_lookup: PathLookup, path: Path) -> MaybeDependency:
        return self._file_resolver.resolve_file(path_lookup, path)

    def register_library(self, path_lookup: PathLookup, *libraries: str) -> list[DependencyProblem]:
        return self._library_resolver.register_library(path_lookup, *libraries)

    def build_local_file_dependency_graph(self, path: Path, session_state: CurrentSessionState) -> MaybeGraph:
        """Builds a dependency graph starting from a file. This method is mainly intended for testing purposes.
        In case of problems, the paths in the problems will be relative to the starting path lookup."""
        maybe = self._file_resolver.resolve_file(self._path_lookup, path)
        if not maybe.dependency:
            return MaybeGraph(None, self._make_relative_paths(maybe.problems, path))
        graph = DependencyGraph(maybe.dependency, None, self, self._path_lookup, session_state)
        container = maybe.dependency.load(graph.path_lookup)
        if container is None:
            problem = DependencyProblem('cannot-load-file', f"Could not load file {path}")
            return MaybeGraph(None, [problem])
        problems = container.build_dependency_graph(graph)
        if problems:
            problems = self._make_relative_paths(problems, path)
        return MaybeGraph(graph, problems)

    def build_notebook_dependency_graph(self, path: Path, session_state: CurrentSessionState) -> MaybeGraph:
        """Builds a dependency graph starting from a notebook. This method is mainly intended for testing purposes.
        In case of problems, the paths in the problems will be relative to the starting path lookup."""
        # the notebook is the root of the graph, so there's no context to inherit
        maybe = self._notebook_resolver.resolve_notebook(self._path_lookup, path, inherit_context=False)
        if not maybe.dependency:
            return MaybeGraph(None, self._make_relative_paths(maybe.problems, path))
        graph = DependencyGraph(maybe.dependency, None, self, self._path_lookup, session_state)
        container = maybe.dependency.load(graph.path_lookup)
        if container is None:
            problem = DependencyProblem('cannot-load-notebook', f"Could not load notebook {path}")
            return MaybeGraph(None, [problem])
        problems = container.build_dependency_graph(graph)
        if problems:
            problems = self._make_relative_paths(problems, path)
        return MaybeGraph(graph, problems)

    def _make_relative_paths(self, problems: list[DependencyProblem], path: Path) -> list[DependencyProblem]:
        adjusted_problems = []
        for problem in problems:
            out_path = path if problem.is_path_missing() else problem.source_path
            if out_path.is_absolute() and out_path.is_relative_to(self._path_lookup.cwd):
                out_path = out_path.relative_to(self._path_lookup.cwd)
            adjusted_problems.append(dataclasses.replace(problem, source_path=out_path))
        return adjusted_problems

    def __repr__(self):
        return f"<DependencyResolver {self._notebook_resolver} {self._import_resolver} {self._file_resolver}, {self._path_lookup}>"


MISSING_SOURCE_PATH = "<MISSING_SOURCE_PATH>"


@dataclass
class DependencyProblem:
    code: str
    message: str
    source_path: Path = Path(MISSING_SOURCE_PATH)
    # Lines and columns are both 0-based: the first line is line 0.
    start_line: int = -1
    start_col: int = -1
    end_line: int = -1
    end_col: int = -1

    def is_path_missing(self):
        return self.source_path == Path(MISSING_SOURCE_PATH)

    def as_advisory(self) -> 'Advisory':
        return Advisory(
            code=self.code,
            message=self.message,
            start_line=self.start_line,
            start_col=self.start_col,
            end_line=self.end_line,
            end_col=self.end_col,
        )

    @staticmethod
    def from_node(code: str, message: str, node: NodeNG) -> DependencyProblem:
        # Astroid line numbers are 1-based.
        return DependencyProblem(
            code=code,
            message=message,
            start_line=(node.lineno or 1) - 1,
            start_col=node.col_offset,
            end_line=(node.end_lineno or 1) - 1,
            end_col=(node.end_col_offset or 0),
        )


@dataclass
class MaybeGraph:
    graph: DependencyGraph | None
    problems: list[DependencyProblem]

    @property
    def failed(self):
        return len(self.problems) > 0


class InheritedContext:

    @classmethod
    def from_route(cls, graph: DependencyGraph, path_lookup: PathLookup, route: list[Dependency]) -> InheritedContext:
        context = InheritedContext(None, False)
        for i, dependency in enumerate(route):
            if i >= len(route) - 1:
                break
            next_path = route[i + 1].path
            container = dependency.load(path_lookup)
            if container is None:
                logger.warning(f"Could not load content of {dependency.path}")
                return context
            local = container.build_inherited_context(graph, next_path)
            # only copy 'found' flag if this is the last parent
            context = context.append(local, i == len(route) - 2)
        return context.finalize()

    def __init__(self, tree: Tree | None, found: bool):
        self._tree = tree
        self._found = found

    @property
    def tree(self):
        return self._tree

    @property
    def found(self):
        return self._found

    def append(self, context: InheritedContext, copy_found: bool) -> InheritedContext:
        # we should never append to a found context
        if self.found:
            raise ValueError("Appending to an already resolved InheritedContext is illegal!")
        tree = context.tree
        found = copy_found and context.found
        if tree is None:
            return InheritedContext(self._tree, found)
        if self._tree is None:
            self._tree = Tree.new_module()
        self._tree.append_tree(context.tree)
        return InheritedContext(self._tree, found)

    def finalize(self) -> InheritedContext:
        # hacky stuff for aligning with Astroid's inference engine behavior
        # the engine checks line numbers to skip variables that are not in scope of the current frame
        # see https://github.com/pylint-dev/astroid/blob/5b665e7e760a7181625a24b3635e9fec7b174d87/astroid/filter_statements.py#L113
        # this is problematic when linting code fragments that refer to inherited code with unrelated line numbers
        # here we fool the engine by pretending that all nodes from context have negative line numbers
        if self._tree is None:
            return self
        tree = self._tree.renumber(-1)
        return InheritedContext(tree, self.found)


T = TypeVar("T")


class DependencyGraphWalker(abc.ABC, Generic[T]):

    def __init__(self, graph: DependencyGraph, walked_paths: set[Path], path_lookup: PathLookup):
        self._graph = graph
        self._walked_paths = walked_paths
        self._path_lookup = path_lookup
        self._lineage: list[Dependency] = []

    def __iter__(self) -> Iterator[T]:
        for dependency in self._graph.root_dependencies:
            # the dependency is a root, so its path is the one to use
            # for computing lineage and building python global context
            root_path = dependency.path
            yield from self._iter_one(dependency, self._graph, root_path)

    def _iter_one(self, dependency: Dependency, graph: DependencyGraph, root_path: Path) -> Iterable[T]:
        if dependency.path in self._walked_paths:
            return
        self._lineage.append(dependency)
        self._walked_paths.add(dependency.path)
        self._log_walk_one(dependency)
        if dependency.path.is_file() or is_a_notebook(dependency.path):
            inherited_tree = graph.root.build_inherited_tree(root_path, dependency.path)
            path_lookup = self._path_lookup.change_directory(dependency.path.parent)
            yield from self._process_dependency(dependency, path_lookup, inherited_tree)
            maybe_graph = graph.locate_dependency(dependency.path)
            # missing graph problems have already been reported while building the graph
            if maybe_graph.graph:
                child_graph = maybe_graph.graph
                for child_dependency in child_graph.local_dependencies:
                    yield from self._iter_one(child_dependency, child_graph, root_path)
        self._lineage.pop()

    def _log_walk_one(self, dependency: Dependency):
        logger.debug(f'Analyzing dependency: {dependency}')

    @abc.abstractmethod
    def _process_dependency(
        self, dependency: Dependency, path_lookup: PathLookup, inherited_tree: Tree | None
    ) -> Iterable[T]: ...

    @property
    def lineage(self) -> list[LineageAtom]:
        lists: list[list[LineageAtom]] = [dependency.lineage for dependency in self._lineage]
        return list(itertools.chain(*lists))



# pylint: disable=invalid-name,unused-argument
import logging

from databricks.labs.blueprint.installation import Installation
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)


def upgrade(installation: Installation, ws: WorkspaceClient):
    installation.upload('logs/README.md', b'# This folder contains logs from UCX workflows')



import logging
from packaging import version

from pylsp import hookimpl  # type: ignore
from pylsp.config.config import Config  # type: ignore
from pylsp.workspace import Document  # type: ignore
from databricks.sdk.service.workspace import Language

from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex, TableMigrationStatus
from databricks.labs.ucx.source_code.base import CurrentSessionState
from databricks.labs.ucx.source_code.linters.context import LinterContext
from databricks.labs.ucx.source_code.lsp import Diagnostic


logger = logging.getLogger(__name__)


@hookimpl
def pylsp_lint(config: Config, document: Document) -> list[dict]:
    cfg = config.plugin_settings('pylsp_ucx', document_path=document.uri)

    migration_index = TableMigrationIndex([TableMigrationStatus.from_json(st) for st in cfg.get('migration_index', [])])

    session_state = CurrentSessionState(
        data_security_mode=CurrentSessionState.parse_security_mode(cfg.get('dataSecurityMode', None)),
        dbr_version=parse_dbr_version(cfg.get('dbrVersion', None)),
        is_serverless=bool(cfg.get('isServerless', False)),
    )
    languages = LinterContext(index=migration_index, session_state=session_state)
    analyser = languages.linter(Language.PYTHON)
    code = document.source
    diagnostics = [Diagnostic.from_advice(_) for _ in analyser.lint(code)]
    return [d.as_dict() for d in diagnostics]


def parse_dbr_version(version_str: str | None) -> tuple[int, int] | None:
    if not version_str:
        return None
    try:
        release_version = version.parse(version_str).release
        return release_version[0], release_version[1]
    except version.InvalidVersion:
        logger.warning(f'Incorrect DBR version string: {version_str}')
        return None



from __future__ import annotations

import codecs
import dataclasses
import locale
import logging
import sys
from abc import abstractmethod, ABC
from collections.abc import Iterable
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any

from astroid import AstroidSyntaxError, NodeNG  # type: ignore
from sqlglot import Expression, parse as parse_sql, ParseError as SqlParseError

from databricks.sdk.service import compute
from databricks.sdk.service.workspace import Language

from databricks.labs.blueprint.paths import WorkspacePath

from databricks.labs.ucx.source_code.python.python_ast import Tree

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

# Code mapping between LSP, PyLint, and our own diagnostics:
# | LSP                       | PyLint     | Our            |
# |---------------------------|------------|----------------|
# | Severity.ERROR            | Error      | Failure()      |
# | Severity.WARN             | Warning    | Advisory()     |
# | DiagnosticTag.DEPRECATED  | Warning    | Deprecation()  |
# | Severity.INFO             | Info       | Advice()       |
# | Severity.HINT             | Convention | Convention()   |
# | DiagnosticTag.UNNECESSARY | Refactor   | Convention()   |

logger = logging.getLogger(__name__)


@dataclass
class Advice:
    code: str
    message: str
    # Lines and columns are both 0-based: the first line is line 0.
    start_line: int
    start_col: int
    end_line: int
    end_col: int

    def as_advisory(self) -> 'Advisory':
        return Advisory(**self.__dict__)

    def as_failure(self) -> 'Failure':
        return Failure(**self.__dict__)

    def as_deprecation(self) -> 'Deprecation':
        return Deprecation(**self.__dict__)

    def as_convention(self) -> 'Convention':
        return Convention(**self.__dict__)

    def for_path(self, path: Path) -> LocatedAdvice:
        return LocatedAdvice(self, path)

    @classmethod
    def from_node(cls, *, code: str, message: str, node: NodeNG) -> Advice:
        # Astroid lines are 1-based.
        return cls(
            code=code,
            message=message,
            start_line=(node.lineno or 1) - 1,
            start_col=node.col_offset or 0,
            end_line=(node.end_lineno or 1) - 1,
            end_col=node.end_col_offset,
        )

    def replace_from_node(self, node: NodeNG) -> Advice:
        # Astroid lines are 1-based.
        return dataclasses.replace(
            self,
            start_line=(node.lineno or 1) - 1,
            start_col=node.col_offset,
            end_line=(node.end_lineno or 1) - 1,
            end_col=node.end_col_offset,
        )


@dataclass
class LocatedAdvice:
    advice: Advice
    path: Path

    @property
    def is_unknown(self):
        return self.path == Path('UNKNOWN')

    def message_relative_to(self, base: Path, *, default: Path | None = None) -> str:
        advice = self.advice
        path = self.path
        if self.is_unknown:
            logger.debug(f'THIS IS A BUG! {advice.code}:{advice.message} has unknown path')
        if default is not None:
            path = default
        try:
            path = path.relative_to(base)
        except ValueError:
            logger.debug(f'Not a relative path: {path} to base: {base}')
        # increment start_line because it is 0-based whereas IDEs are usually 1-based
        return f"./{path.as_posix()}:{advice.start_line+1}:{advice.start_col}: [{advice.code}] {advice.message}"


class Advisory(Advice):
    """A warning that does not prevent the code from running."""


class Failure(Advice):
    """An error that prevents the code from running."""


class Deprecation(Advice):
    """An advisory that suggests to replace the code with a newer version."""


class Convention(Advice):
    """A suggestion for a better way to write the code."""


class Linter:
    @abstractmethod
    def lint(self, code: str) -> Iterable[Advice]: ...


class SqlLinter(Linter):

    def lint(self, code: str) -> Iterable[Advice]:
        try:
            expressions = parse_sql(code, read='databricks')
            for expression in expressions:
                if not expression:
                    continue
                yield from self.lint_expression(expression)
        except SqlParseError as e:
            logger.debug(f"Failed to parse SQL: {code}", exc_info=e)
            yield self.sql_parse_failure(code)

    @staticmethod
    def sql_parse_failure(code: str):
        return Failure(
            code='sql-parse-error',
            message=f"SQL expression is not supported yet: {code}",
            # SQLGlot does not propagate tokens yet. See https://github.com/tobymao/sqlglot/issues/3159
            start_line=0,
            start_col=0,
            end_line=0,
            end_col=1024,
        )

    @abstractmethod
    def lint_expression(self, expression: Expression) -> Iterable[Advice]: ...


class PythonLinter(Linter):

    def lint(self, code: str) -> Iterable[Advice]:
        tree = Tree.normalize_and_parse(code)
        yield from self.lint_tree(tree)

    @abstractmethod
    def lint_tree(self, tree: Tree) -> Iterable[Advice]: ...


class Fixer(ABC):

    @property
    @abstractmethod
    def name(self) -> str: ...

    @abstractmethod
    def apply(self, code: str) -> str: ...


@dataclass
class LineageAtom:

    object_type: str
    object_id: str
    other: dict[str, str] | None = None


@dataclass
class SourceInfo:

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        source_lineage = data.get("source_lineage", None)
        if isinstance(source_lineage, list) and len(source_lineage) > 0 and isinstance(source_lineage[0], dict):
            lineage_atoms = [LineageAtom(**lineage) for lineage in source_lineage]
            data["source_lineage"] = lineage_atoms
        return cls(**data)

    UNKNOWN = "unknown"

    source_id: str = UNKNOWN
    source_timestamp: datetime = datetime.fromtimestamp(0)
    source_lineage: list[LineageAtom] = field(default_factory=list)
    assessment_start_timestamp: datetime = datetime.fromtimestamp(0)
    assessment_end_timestamp: datetime = datetime.fromtimestamp(0)

    def replace_source(
        self,
        source_id: str | None = None,
        source_lineage: list[LineageAtom] | None = None,
        source_timestamp: datetime | None = None,
    ):
        return dataclasses.replace(
            self,
            source_id=source_id or self.source_id,
            source_timestamp=source_timestamp or self.source_timestamp,
            source_lineage=source_lineage or self.source_lineage,
        )

    def replace_assessment_infos(
        self, assessment_start: datetime | None = None, assessment_end: datetime | None = None
    ):
        return dataclasses.replace(
            self,
            assessment_start_timestamp=assessment_start or self.assessment_start_timestamp,
            assessment_end_timestamp=assessment_end or self.assessment_end_timestamp,
        )


@dataclass
class UsedTable(SourceInfo):

    @classmethod
    def parse(cls, value: str, default_schema: str) -> UsedTable:
        parts = value.split(".")
        if len(parts) >= 3:
            catalog_name = parts.pop(0)
        else:
            catalog_name = "hive_metastore"
        if len(parts) >= 2:
            schema_name = parts.pop(0)
        else:
            schema_name = default_schema
        return UsedTable(catalog_name=catalog_name, schema_name=schema_name, table_name=parts[0])

    catalog_name: str = SourceInfo.UNKNOWN
    schema_name: str = SourceInfo.UNKNOWN
    table_name: str = SourceInfo.UNKNOWN
    is_read: bool = True
    is_write: bool = False


class TableCollector(ABC):

    @abstractmethod
    def collect_tables(self, source_code: str) -> Iterable[UsedTable]: ...


@dataclass
class TableInfoNode:
    table: UsedTable
    node: NodeNG


class TablePyCollector(TableCollector, ABC):

    def collect_tables(self, source_code: str):
        tree = Tree.normalize_and_parse(source_code)
        for table_node in self.collect_tables_from_tree(tree):
            yield table_node.table

    @abstractmethod
    def collect_tables_from_tree(self, tree: Tree) -> Iterable[TableInfoNode]: ...


class TableSqlCollector(TableCollector, ABC): ...


@dataclass
class DirectFsAccess(SourceInfo):
    """A record describing a Direct File System Access"""

    path: str = SourceInfo.UNKNOWN
    is_read: bool = False
    is_write: bool = False


@dataclass
class DirectFsAccessNode:
    dfsa: DirectFsAccess
    node: NodeNG


class DfsaCollector(ABC):

    @abstractmethod
    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]: ...


class DfsaPyCollector(DfsaCollector, ABC):

    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]:
        tree = Tree.normalize_and_parse(source_code)
        for dfsa_node in self.collect_dfsas_from_tree(tree):
            yield dfsa_node.dfsa

    @abstractmethod
    def collect_dfsas_from_tree(self, tree: Tree) -> Iterable[DirectFsAccessNode]: ...


class DfsaSqlCollector(DfsaCollector, ABC): ...


# The default schema to use when the schema is not specified in a table reference
# See: https://spark.apache.org/docs/3.0.0-preview/sql-ref-syntax-qry-select-usedb.html
DEFAULT_CATALOG = 'hive_metastore'
DEFAULT_SCHEMA = 'default'


@dataclass
class CurrentSessionState:
    """
    A data class that represents the current state of a session.

    This class can be used to track various aspects of a session, such as the current schema.

    Attributes:
        catalog (str): The current schema of the session. If not provided, it defaults to 'DEFAULT_CATALOG'.
        schema (str): The current schema of the session. If not provided, it defaults to 'DEFAULT_SCHEMA'.
    """

    schema: str = DEFAULT_SCHEMA
    catalog: str = DEFAULT_CATALOG
    spark_conf: dict[str, str] | None = None
    named_parameters: dict[str, str] | None = None
    data_security_mode: compute.DataSecurityMode | None = None
    is_serverless: bool = False
    dbr_version: tuple[int, int] | None = None

    @classmethod
    def from_json(cls, json: dict) -> CurrentSessionState:
        return cls(
            schema=json.get('schema', DEFAULT_SCHEMA),
            catalog=json.get('catalog', DEFAULT_CATALOG),
            spark_conf=json.get('spark_conf', None),
            named_parameters=json.get('named_parameters', None),
            data_security_mode=cls.parse_security_mode(json.get('data_security_mode', None)),
            is_serverless=json.get('is_serverless', False),
            dbr_version=tuple(json['dbr_version']) if 'dbr_version' in json else None,
        )

    @staticmethod
    def parse_security_mode(mode_str: str | None) -> compute.DataSecurityMode | None:
        try:
            return compute.DataSecurityMode(mode_str) if mode_str else None
        except ValueError:
            logger.warning(f'Unknown data_security_mode {mode_str}')
            return None


class SqlSequentialLinter(SqlLinter, DfsaCollector, TableCollector):

    def __init__(
        self,
        linters: list[SqlLinter],
        dfsa_collectors: list[DfsaSqlCollector],
        table_collectors: list[TableSqlCollector],
    ):
        self._linters = linters
        self._dfsa_collectors = dfsa_collectors
        self._table_collectors = table_collectors

    def lint_expression(self, expression: Expression) -> Iterable[Advice]:
        for linter in self._linters:
            yield from linter.lint_expression(expression)

    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]:
        for collector in self._dfsa_collectors:
            yield from collector.collect_dfsas(source_code)

    def collect_tables(self, source_code: str) -> Iterable[UsedTable]:
        for collector in self._table_collectors:
            yield from collector.collect_tables(source_code)


class PythonSequentialLinter(Linter, DfsaCollector, TableCollector):

    def __init__(
        self,
        linters: list[PythonLinter],
        dfsa_collectors: list[DfsaPyCollector],
        table_collectors: list[TablePyCollector],
    ):
        self._linters = linters
        self._dfsa_collectors = dfsa_collectors
        self._table_collectors = table_collectors
        self._tree: Tree | None = None

    def lint(self, code: str) -> Iterable[Advice]:
        try:
            tree = self._parse_and_append(code)
            yield from self.lint_tree(tree)
        except AstroidSyntaxError as e:
            yield Failure('syntax-error', str(e), 0, 0, 0, 0)

    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
        for linter in self._linters:
            yield from linter.lint_tree(tree)

    def _parse_and_append(self, code: str) -> Tree:
        tree = Tree.normalize_and_parse(code)
        self.append_tree(tree)
        return tree

    def append_tree(self, tree: Tree):
        self._make_tree().append_tree(tree)

    def append_nodes(self, nodes: list[NodeNG]):
        self._make_tree().append_nodes(nodes)

    def append_globals(self, globs: dict):
        self._make_tree().append_globals(globs)

    def process_child_cell(self, code: str):
        try:
            this_tree = self._make_tree()
            tree = Tree.normalize_and_parse(code)
            this_tree.append_tree(tree)
        except AstroidSyntaxError as e:
            # error already reported when linting enclosing notebook
            logger.warning(f"Failed to parse Python cell: {code}", exc_info=e)

    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]:
        try:
            tree = self._parse_and_append(source_code)
            for dfsa_node in self.collect_dfsas_from_tree(tree):
                yield dfsa_node.dfsa
        except AstroidSyntaxError as e:
            logger.warning('syntax-error', exc_info=e)

    def collect_dfsas_from_tree(self, tree: Tree) -> Iterable[DirectFsAccessNode]:
        for collector in self._dfsa_collectors:
            yield from collector.collect_dfsas_from_tree(tree)

    def collect_tables(self, source_code: str) -> Iterable[UsedTable]:
        try:
            tree = self._parse_and_append(source_code)
            for table_node in self.collect_tables_from_tree(tree):
                yield table_node.table
        except AstroidSyntaxError as e:
            logger.warning('syntax-error', exc_info=e)

    def collect_tables_from_tree(self, tree: Tree) -> Iterable[TableInfoNode]:
        for collector in self._table_collectors:
            yield from collector.collect_tables_from_tree(tree)

    def _make_tree(self) -> Tree:
        if self._tree is None:
            self._tree = Tree.new_module()
        return self._tree


SUPPORTED_EXTENSION_LANGUAGES = {
    '.py': Language.PYTHON,
    '.sql': Language.SQL,
}


def file_language(path: Path) -> Language | None:
    return SUPPORTED_EXTENSION_LANGUAGES.get(path.suffix.lower())


def guess_encoding(path: Path):
    # some files encode a unicode BOM (byte-order-mark), so let's use that if available
    with path.open('rb') as _file:
        raw = _file.read(4)
        if raw.startswith(codecs.BOM_UTF32_LE) or raw.startswith(codecs.BOM_UTF32_BE):
            return 'utf-32'
        if raw.startswith(codecs.BOM_UTF16_LE) or raw.startswith(codecs.BOM_UTF16_BE):
            return 'utf-16'
        if raw.startswith(codecs.BOM_UTF8):
            return 'utf-8-sig'
        # no BOM, let's use default encoding
        return locale.getpreferredencoding(False)


# duplicated from CellLanguage to prevent cyclic import
LANGUAGE_COMMENT_PREFIXES = {Language.PYTHON: '#', Language.SCALA: '//', Language.SQL: '--'}
NOTEBOOK_HEADER = "Databricks notebook source"


def is_a_notebook(path: Path, content: str | None = None) -> bool:
    if isinstance(path, WorkspacePath):
        return path.is_notebook()
    if not path.is_file():
        return False
    language = file_language(path)
    if not language:
        return False
    magic_header = f"{LANGUAGE_COMMENT_PREFIXES.get(language)} {NOTEBOOK_HEADER}"
    if content is not None:
        return content.startswith(magic_header)
    try:
        with path.open('rt', encoding=guess_encoding(path)) as f:
            file_header = f.read(len(magic_header))
    except (FileNotFoundError, UnicodeDecodeError, PermissionError):
        logger.warning(f"Could not read file {path}")
        return False
    return file_header == magic_header



import logging

from databricks.labs.lsql.backends import SqlBackend
from databricks.labs.lsql.deployment import SchemaDeployer
from databricks.labs.ucx.progress.workflow_runs import WorkflowRun

logger = logging.getLogger(__name__)


class ProgressTrackingInstallation:
    """Install resources for UCX's progress tracking."""

    _SCHEMA = "multiworkspace"

    def __init__(self, sql_backend: SqlBackend, ucx_catalog: str) -> None:
        # `mod` is a required parameter, though, it's not used in this context without views.
        self._schema_deployer = SchemaDeployer(sql_backend, self._SCHEMA, mod=None, catalog=ucx_catalog)

    def run(self) -> None:
        self._schema_deployer.deploy_schema()
        self._schema_deployer.deploy_table("workflow_runs", WorkflowRun)
        logger.info("Installation completed successfully!")



import logging
from collections.abc import Callable, Iterable, Iterator
from typing import TypeVar

from sqlglot import parse, ParseError
from sqlglot.expressions import Create, Delete, Drop, Expression, Select, Table, Update, Use

from databricks.labs.ucx.source_code.base import UsedTable, CurrentSessionState

logger = logging.getLogger(__name__)

T = TypeVar("T")
E = TypeVar("E", bound=Expression)


class SqlExpression:

    def __init__(self, expression: Expression):
        self._expression = expression

    def collect_table_infos(self, required_catalog: str, session_state: CurrentSessionState) -> Iterable[UsedTable]:
        for table in self._expression.find_all(Table):
            info = self._collect_table_info(table, required_catalog, session_state)
            if info:
                yield info

    def _collect_table_info(
        self, table: Table, required_catalog: str, session_state: CurrentSessionState
    ) -> UsedTable | None:
        if isinstance(self._expression, Use):
            # Sqlglot captures the database name in the Use statement as a Table, with
            # the schema  as the table name.
            session_state.schema = table.name
            return None
        if isinstance(self._expression, Drop) and getattr(self._expression, "kind", None) == "SCHEMA":
            # Sqlglot captures the schema name in the Drop statement as a Table, with
            # the schema  as the db name.
            return None
        if isinstance(self._expression, Create) and getattr(self._expression, "kind", None) == "SCHEMA":
            # Sqlglot captures the schema name in the Create statement as a Table, with
            # the schema  as the db name.
            session_state.schema = table.db
            return None
        # we only return tables in the required catalog
        catalog_name = table.catalog or required_catalog
        if catalog_name != required_catalog:
            return None
        # Sqlglot uses db instead of schema, watch out for that
        src_schema = table.db if table.db else session_state.schema
        if not src_schema:
            logger.error(f"Could not determine schema for table {table.name}")
            return None
        return UsedTable(
            catalog_name=catalog_name,
            schema_name=src_schema,
            table_name=table.name,
            is_read=isinstance(self._expression, Select),
            is_write=isinstance(self._expression, (Create, Update, Delete)),
        )

    def find_all(self, klass: type[E]) -> Iterator[E]:
        return self._expression.find_all(klass)


class SqlParser:

    @classmethod
    def walk_expressions(cls, sql_code: str, callback: Callable[[SqlExpression], Iterable[T]]) -> Iterable[T]:
        try:
            expressions = parse(sql_code, read='databricks')
            for expression in expressions:
                if not expression:
                    continue
                yield from callback(SqlExpression(expression))
        except ParseError as e:
            logger.debug(f"Failed to parse SQL: {sql_code}", exc_info=e)
            raise e



from __future__ import annotations

import dataclasses
import logging
from collections.abc import Iterable
from pathlib import Path
from typing import cast

from astroid import AstroidSyntaxError, ImportFrom, Try, Name  # type: ignore

from databricks.labs.ucx.source_code.graph import (
    DependencyGraphContext,
    DependencyProblem,
    InheritedContext,
)
from databricks.labs.ucx.source_code.linters.imports import (
    SysPathChange,
    DbutilsPyLinter,
    ImportSource,
    NotebookRunCall,
    UnresolvedPath,
)
from databricks.labs.ucx.source_code.notebooks.magic import MagicLine
from databricks.labs.ucx.source_code.python.python_ast import Tree, NodeBase

logger = logging.getLogger(__name__)


class PythonCodeAnalyzer:

    def __init__(self, context: DependencyGraphContext, python_code: str):
        self._context = context
        self._python_code = python_code

    def build_graph(self) -> list[DependencyProblem]:
        """Check python code for dependency-related problems.

        Returns:
            A list of dependency problems; position information is relative to the python code itself.
        """
        problems: list[DependencyProblem] = []
        try:
            _, nodes, parse_problems = self._parse_and_extract_nodes()
            problems.extend(parse_problems)
        except AstroidSyntaxError as e:
            logger.debug(f"Could not parse Python code: {self._python_code}", exc_info=True)
            problems.append(DependencyProblem('parse-error', f"Could not parse Python code: {e}"))
            return problems
        for base_node in nodes:
            for problem in self._build_graph_from_node(base_node):
                # Astroid line numbers are 1-based.
                problem = dataclasses.replace(
                    problem,
                    start_line=base_node.node.lineno - 1,
                    start_col=base_node.node.col_offset,
                    end_line=(base_node.node.end_lineno or 1) - 1,
                    end_col=base_node.node.end_col_offset or 0,
                )
                problems.append(problem)
        return problems

    def build_inherited_context(self, child_path: Path) -> InheritedContext:
        try:
            tree, nodes, _ = self._parse_and_extract_nodes()
        except AstroidSyntaxError:
            logger.debug(f"Could not parse Python code: {self._python_code}", exc_info=True)
            return InheritedContext(None, False)
        if len(nodes) == 0:
            return InheritedContext(tree, False)
        context = InheritedContext(Tree.new_module(), False)
        last_line = -1
        for base_node in nodes:
            # append nodes
            node_line = base_node.node.lineno
            nodes = tree.nodes_between(last_line + 1, node_line - 1)
            context.tree.append_nodes(nodes)
            globs = tree.globals_between(last_line + 1, node_line - 1)
            context.tree.append_globals(globs)
            last_line = node_line
            # process node
            child_context = self._build_inherited_context_from_node(base_node, child_path)
            context = context.append(child_context, True)
            if context.found:
                return context
        line_count = tree.line_count()
        if last_line < line_count:
            nodes = tree.nodes_between(last_line + 1, line_count)
            context.tree.append_nodes(nodes)
            globs = tree.globals_between(last_line + 1, line_count)
            context.tree.append_globals(globs)
        return context

    def _parse_and_extract_nodes(self) -> tuple[Tree, list[NodeBase], Iterable[DependencyProblem]]:
        problems: list[DependencyProblem] = []
        tree = Tree.normalize_and_parse(self._python_code)
        syspath_changes = SysPathChange.extract_from_tree(self._context.session_state, tree)
        run_calls = DbutilsPyLinter.list_dbutils_notebook_run_calls(tree)
        import_sources: list[ImportSource]
        import_problems: list[DependencyProblem]
        import_sources, import_problems = ImportSource.extract_from_tree(tree, DependencyProblem.from_node)
        problems.extend(import_problems)
        magic_lines, command_problems = MagicLine.extract_from_tree(tree, DependencyProblem.from_node)
        problems.extend(command_problems)
        # need to evaluate things in intertwined sequence so concat and sort them
        nodes: list[NodeBase] = cast(list[NodeBase], syspath_changes + run_calls + import_sources + magic_lines)
        nodes = sorted(nodes, key=lambda node: (node.node.lineno, node.node.col_offset))
        return tree, nodes, problems

    def _build_graph_from_node(self, base_node: NodeBase) -> Iterable[DependencyProblem]:
        if isinstance(base_node, SysPathChange):
            yield from self._mutate_path_lookup(base_node)
        elif isinstance(base_node, NotebookRunCall):
            yield from self._register_notebook(base_node)
        elif isinstance(base_node, ImportSource):
            yield from self._register_import(base_node)
        elif isinstance(base_node, MagicLine):
            yield from base_node.build_dependency_graph(self._context.parent)
        else:
            logger.warning(f"Can't build graph for node {NodeBase.__name__} of type {type(base_node).__name__}")

    def _build_inherited_context_from_node(self, base_node: NodeBase, child_path: Path) -> InheritedContext:
        if isinstance(base_node, SysPathChange):
            self._mutate_path_lookup(base_node)
            return InheritedContext(None, False)
        if isinstance(base_node, ImportSource):
            # nothing to do, Astroid takes care of imports
            return InheritedContext(None, False)
        if isinstance(base_node, NotebookRunCall):
            # nothing to do, dbutils.notebook.run uses a dedicated context
            return InheritedContext(None, False)
        if isinstance(base_node, MagicLine):
            return base_node.build_inherited_context(self._context, child_path)
        logger.warning(f"Can't build inherited context for node {NodeBase.__name__} of type {type(base_node).__name__}")
        return InheritedContext(None, False)

    def _register_import(self, base_node: ImportSource) -> Iterable[DependencyProblem]:
        prefix = ""
        if isinstance(base_node.node, ImportFrom) and base_node.node.level is not None:
            prefix = "." * base_node.node.level
        name = base_node.name or ""
        problems = self._context.parent.register_import(prefix + name)
        for problem in problems:
            prob = self._filter_import_problem_in_try_except(problem, base_node)
            if prob is not None:
                yield prob

    @classmethod
    def _filter_import_problem_in_try_except(
        cls, problem: DependencyProblem, base_node: ImportSource
    ) -> DependencyProblem | None:
        if problem.code != 'import-not-found':
            return problem
        # is base_node in a try-except clause ?
        node = base_node.node.parent
        while node and not isinstance(node, Try):
            node = node.parent
        if cls._is_try_except_import_error(node):
            return None
        return problem

    @classmethod
    def _is_try_except_import_error(cls, node: Try | None) -> bool:
        if not isinstance(node, Try):
            return False
        for handler in node.handlers:
            if isinstance(handler.type, Name):
                if handler.type.name == "ImportError":
                    return True
        return False

    def _register_notebook(self, base_node: NotebookRunCall) -> Iterable[DependencyProblem]:
        has_unresolved, paths = base_node.get_notebook_paths(self._context.session_state)
        if has_unresolved:
            yield DependencyProblem(
                'dependency-cannot-compute-value',
                f"Can't check dependency from {base_node.node.as_string()} because the expression cannot be computed",
            )
        for path in paths:
            # notebooks ran via dbutils.notebook.run do not inherit or propagate context
            yield from self._context.parent.register_notebook(Path(path), False)

    def _mutate_path_lookup(self, change: SysPathChange) -> Iterable[DependencyProblem]:
        if isinstance(change, UnresolvedPath):
            yield DependencyProblem(
                'sys-path-cannot-compute-value',
                f"Can't update sys.path from {change.node.as_string()} because the expression cannot be computed",
            )
            return
        change.apply_to(self._context.path_lookup)



import datetime as dt
import logging
from dataclasses import dataclass

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import NotFound


logger = logging.getLogger(__name__)


@dataclass(frozen=True, kw_only=True)
class WorkflowRun:
    started_at: dt.datetime
    """The timestamp of the workflow run start."""

    finished_at: dt.datetime
    """The timestamp of the workflow run end."""

    workspace_id: int
    """The workspace id in which the workflow ran."""

    workflow_name: str
    """The workflow name that ran."""

    workflow_id: int
    """"The workflow id of the workflow that ran."""

    workflow_run_id: int
    """The workflow run id."""

    workflow_run_attempt: int
    """The workflow run attempt."""


class WorkflowRunRecorder:
    """Record workflow runs in a database."""

    def __init__(
        self,
        sql_backend: SqlBackend,
        ucx_catalog: str,
        *,
        workspace_id: int,
        workflow_name: str,
        workflow_id: int,
        workflow_run_id: int,
        workflow_run_attempt: int,
        workflow_start_time: str,
    ):
        self._sql_backend = sql_backend
        self._full_table_name = f"{ucx_catalog}.multiworkspace.workflow_runs"
        self._workspace_id = workspace_id
        self._workflow_name = workflow_name
        self._workflow_start_time = workflow_start_time
        self._workflow_id = workflow_id
        self._workflow_run_id = workflow_run_id
        self._workflow_run_attempt = workflow_run_attempt

    def record(self) -> None:
        """Record a workflow run."""
        workflow_run = WorkflowRun(
            started_at=dt.datetime.fromisoformat(self._workflow_start_time),
            finished_at=dt.datetime.now(tz=dt.timezone.utc).replace(microsecond=0),
            workspace_id=self._workspace_id,
            workflow_name=self._workflow_name,
            workflow_id=self._workflow_id,
            workflow_run_id=self._workflow_run_id,
            workflow_run_attempt=self._workflow_run_attempt,
        )
        try:
            self._sql_backend.save_table(
                self._full_table_name,
                [workflow_run],
                WorkflowRun,
                mode="append",
            )
        except NotFound as e:
            logger.error(f"Workflow run table not found: {self._full_table_name}", exc_info=e)



from dataclasses import dataclass

from databricks.sdk.core import Config

__all__ = ["WorkspaceConfig"]


@dataclass
class WorkspaceConfig:  # pylint: disable=too-many-instance-attributes
    __file__ = "config.yml"
    __version__ = 2

    inventory_database: str
    ucx_catalog: str = "ucx"  # Catalog to store UCX artifact tables (shared across workspaces)
    # Group name conversion parameters.
    workspace_group_regex: str | None = None
    workspace_group_replace: str | None = None
    account_group_regex: str | None = None
    group_match_by_external_id: bool = False
    # Includes group names for migration. If not specified, all matching groups will be picked up
    include_group_names: list[str] | None = None
    renamed_group_prefix: str | None = "db-temp-"
    instance_pool_id: str | None = None
    # in v3, warehouse_id should be part of connect
    warehouse_id: str | None = None
    connect: Config | None = None
    num_threads: int | None = 10
    database_to_catalog_mapping: dict[str, str] | None = None
    default_catalog: str | None = "ucx_default"  # DEPRECATED: Keeping to avoid errors when loading old configurations
    log_level: str | None = "INFO"

    # Starting path for notebooks and directories crawler
    workspace_start_path: str = "/"
    instance_profile: str | None = None
    spark_conf: dict[str, str] | None = None
    min_workers: int | None = 1
    max_workers: int | None = 10

    override_clusters: dict[str, str] | None = None
    policy_id: str | None = None
    num_days_submit_runs_history: int = 30
    uber_spn_id: str | None = None
    uber_instance_profile: str | None = None

    is_terraform_used: bool = False  # Not used, keep for backwards compatibility

    # Whether the assessment should capture a specific list of databases, if not specified, it will list all databases.
    include_databases: list[str] | None = None

    # Whether the tables in mounts crawler should crawl a specific list of mounts.
    # If not specified, it will list all mounts.
    include_mounts: list[str] | None = None
    exclude_paths_in_mount: list[str] | None = None
    include_paths_in_mount: list[str] | None = None

    # Used for limiting the number of jobs to be analysed
    include_job_ids: list[int] | None = None

    # Whether to trigger assessment job after installation
    trigger_job: bool = False

    # List of workspace ids ucx is installed on, only applied to account-level installation
    installed_workspace_ids: list[int] | None = None

    # Threshold for row count comparison during data reconciliation, in percentage
    recon_tolerance_percent: int = 5

    # Whether to upload dependent libraries to the workspace
    upload_dependencies: bool = False

    # [INTERNAL ONLY] Whether the assessment should capture only specific object permissions.
    include_object_permissions: list[str] | None = None

    # [INTERNAL ONLY] Whether the assessment should lint only specific dashboards.
    include_dashboard_ids: list[str] | None = None

    def replace_inventory_variable(self, text: str) -> str:
        return text.replace("$inventory", f"hive_metastore.{self.inventory_database}")

    @classmethod
    def v1_migrate(cls, raw: dict) -> dict:
        # See https://github.com/databrickslabs/ucx/blob/v0.5.0/src/databricks/labs/ucx/config.py#L16-L32
        groups = raw.pop("groups", {})
        selected_groups = groups.get("selected", [])
        if selected_groups:
            raw["include_group_names"] = selected_groups
        raw["renamed_group_prefix"] = groups.get("backup_group_prefix", "db-temp-")
        raw["version"] = 2
        return raw



from databricks.labs.ucx.contexts.workflow_task import RuntimeContext
from databricks.labs.ucx.framework.tasks import Workflow, job_task


class MigrationProgress(Workflow):
    """Experimental workflow that rescans the environment to reflect and track progress that has been made.

    This is a subset of the assessment workflow and covers:

     - Clusters
     - Grants
     - Jobs
     - Pipelines
     - Policies
     - Tables
     - TableMigrationStatus
     - UDFs
    """

    def __init__(self) -> None:
        super().__init__('migration-progress-experimental')

    @job_task
    def crawl_tables(self, ctx: RuntimeContext) -> None:
        """Iterates over all tables in the Hive Metastore of the current workspace and persists their metadata, such
        as _database name_, _table name_, _table type_, _table location_, etc., in the table named
        `$inventory_database.tables`. The metadata stored is then used in the subsequent tasks and workflows to, for
        example, find all Hive Metastore tables that cannot easily be migrated to Unity Catalog."""
        ctx.tables_crawler.snapshot(force_refresh=True)

    @job_task
    def crawl_udfs(self, ctx: RuntimeContext) -> None:
        """Iterates over all UDFs in the Hive Metastore of the current workspace and persists their metadata in the
        table named `$inventory_database.udfs`. This inventory is currently used when scanning securable objects for
        issues with grants that cannot be migrated to Unit Catalog."""
        ctx.udfs_crawler.snapshot(force_refresh=True)

    @job_task(job_cluster="tacl")
    def setup_tacl(self, ctx: RuntimeContext) -> None:
        """(Optimization) Starts `tacl` job cluster in parallel to crawling tables."""

    @job_task(depends_on=[crawl_tables, crawl_udfs, setup_tacl], job_cluster="tacl")
    def crawl_grants(self, ctx: RuntimeContext) -> None:
        """Scans all securable objects for permissions that have been assigned: this include database-level permissions,
        as well permissions directly configured on objects in the (already gathered) table and UDF inventories. The
        captured information is stored in the `$inventory_database.grants` inventory table for further use during the
        migration of legacy ACLs to Unity Catalog permissions.

        Note: This job runs on a separate cluster (named `tacl`) as it requires the proper configuration to have the Table
        ACLs enabled and available for retrieval."""
        ctx.grants_crawler.snapshot(force_refresh=True)

    @job_task
    def assess_jobs(self, ctx: RuntimeContext) -> None:
        """Scans through all the jobs and identifies those that are not compatible with UC. The list of all the jobs is
        stored in the `$inventory.jobs` table.

        It looks for:
          - Clusters with Databricks Runtime (DBR) version earlier than 11.3
          - Clusters using Passthrough Authentication
          - Clusters with incompatible Spark config tags
          - Clusters referencing DBFS locations in one or more config options
        """
        ctx.jobs_crawler.snapshot(force_refresh=True)

    @job_task
    def assess_clusters(self, ctx: RuntimeContext) -> None:
        """Scan through all the clusters and identifies those that are not compatible with UC. The list of all the clusters
        is stored in the`$inventory.clusters` table.

        It looks for:
          - Clusters with Databricks Runtime (DBR) version earlier than 11.3
          - Clusters using Passthrough Authentication
          - Clusters with incompatible spark config tags
          - Clusters referencing DBFS locations in one or more config options
        """
        ctx.clusters_crawler.snapshot(force_refresh=True)

    @job_task
    def assess_pipelines(self, ctx: RuntimeContext) -> None:
        """This module scans through all the Pipelines and identifies those pipelines which has Azure Service Principals
        embedded (who has been given access to the Azure storage accounts via spark configurations) in the pipeline
        configurations.

        It looks for:
          - all the pipelines which has Azure Service Principal embedded in the pipeline configuration

        Subsequently, a list of all the pipelines with matching configurations are stored in the
        `$inventory.pipelines` table."""
        ctx.pipelines_crawler.snapshot(force_refresh=True)

    @job_task
    def crawl_cluster_policies(self, ctx: RuntimeContext) -> None:
        """This module scans through all the Cluster Policies and get the necessary information

        It looks for:
          - Clusters Policies with Databricks Runtime (DBR) version earlier than 11.3

          Subsequently, a list of all the policies with matching configurations are stored in the
        `$inventory.policies` table."""
        ctx.policies_crawler.snapshot(force_refresh=True)

    @job_task(job_cluster="table_migration")
    def setup_table_migration(self, ctx: RuntimeContext) -> None:
        """(Optimization) Starts `table_migration` job cluster in parallel to crawling tables."""

    @job_task(depends_on=[crawl_tables, setup_table_migration], job_cluster="table_migration")
    def refresh_table_migration_status(self, ctx: RuntimeContext) -> None:
        """Scan the tables (and views) in the inventory and record whether each has been migrated or not.

        The results of the scan are stored in the `$inventory.migration_status` inventory table.
        """
        ctx.migration_status_refresher.snapshot(force_refresh=True)

    @job_task(
        depends_on=[
            crawl_grants,
            assess_jobs,
            assess_clusters,
            assess_pipelines,
            crawl_cluster_policies,
            refresh_table_migration_status,
        ]
    )
    def record_workflow_run(self, ctx: RuntimeContext) -> None:
        """Record the workflow run of this workflow."""
        ctx.workflow_run_recorder.record()
